'''
This is the main Lynxy file which provides the functions for public use.
A majority of the code is in the comm.py file, and Lynxy just extends the necessary functions.
The rest can be found in the "_comm" class object.
I know this is probably a clunky way to do this but, if it's not broken then don't fix it, right?
'''

# files
from .comm import Comm as _Comm
from .constants import Constants
from .pool import Pool
from .exceptions import Exceptions

####################################################

# the main class for the keeping everything together
class Lynxy:
    def __init__(self, host: tuple[str, int] = ['', 56774], bind: bool = False):
        '''
        This is the Lynxy class, which creates a Lynxy client object. This allows you to communicate with
        other Lynxy clients and allows you to exchange data between machines. More specifics about certain functionalities
        can be found on the official documentation on Github!

        host: tuple[str, int] = ['', 56774]
        - this is the information for the host machine to bind onto, the information being the IP of the host machine and the chosen
          port. The IP should be left empty, as Lynxy will naturally find out the correct IP. However, the port can be set to whatever you
          desire. This information can be acquired with the get_host function

          >>> client = Lynxy(['', 50004])
          >>> client.get_host() -> ['192.168.68.x', 50004]

        bind: bool = False
        - this decides whether or not you immediately bind to the host IP and port, or if you want to wait until you call the connect function.
        '''
        self._comm = _Comm(host, bind)


    # this gets the host information
    def get_host(self) -> tuple[str, int]: 
        '''
        this function gets the IP and port that the host machine is binded to. This is helpful information for the other machine to use for connecting
        the two clients. 

        The first entry in the tuple is the IP, and the second entry is the port.
        '''
        return self._comm.get_host()


    # this gets the target info
    def get_actual_target(self) -> tuple[str, int]: 
        '''
        when connecting to the other machine, depending on how Lynxy manages things, the other machines port might change when a connection is established.
        You can always check what the current IP and port of the target is by calling this function, which returns a tuple where the first entry is the IP,
        and the second entry is the port.
        '''
        return self._comm.get_actual_target()


    # this function configures heartbeat things for the client
    def config_heartbeat(self, inactive_delay: int = 60, probe_interval: int = 10, probe_count: int = 5) -> None:
        '''
        Lynxy uses heartbeat requests to keep the connection alive, even if no data is sent. 

        inactive_delay: int = 60
        - this represents how long (in integer seconds) Lynxy should wait before sending heartbeat probe to make sure the other machine
            is alive.
        
        probe_interval: int = 10
        - this represents how long (in integer seconds) Lynxy should wait between each heartbeat probe.

        probe_count: int = 5
        - this represents how many heartbeat probes Lynxy should send before terminating the connection due to no response. 

        Do note that on Windows machines, the probe_count is decided by the system.
        '''
        self._comm.config_heartbeat(inactive_delay, probe_interval, probe_count)


    # this function sets behaviors for when connection is lost
    def set_connection(self, connectionType: Constants.ConnectionType) -> None:
        '''
        The connection type determines what Lynxy will do when a closing event happens. There are three types of connection types.

        ConnectionType.EVENT
        - Lynxy will trigger the Event.ON_CLOSE event when the connection is closed. This is the default setting.

        ConnectionType.ERROR
        - Lynxy will raise the error that occurs on connection closing (graceful or not).

        ConnectionType.NONE
        - Lynxy will not do anything when a connection closes.
        '''
        # filter out invalid types
        if type(connectionType) != Constants.ConnectionType: raise TypeError('Invalid connection type')
        # set connection type
        self._comm.connectionType = connectionType
        return None


    # this function connects to the other machine
    def connect(self, 
                target: tuple[str, int], 
                start_recv: bool = True, 
                timeout: int = 10,
                attempts: int = 6,
                connection_bias: Constants.ConnectionBias = Constants.ConnectionBias.NONE
                ) -> None: 
        '''
        This function connects client to the other machine, and exchanges some data to keep things secure.

        target: tuple[str, int]
        - the information of the target machine, the first entry being the IP and the second entry being the port.

        start_recv: bool = True
        - whether to start the thread for recieving or not. If you want to control when you start recieving, set this to False and call on
          the recv function when ready.

        timeout: int = 10
        - how long Lynxy should wait on each attempt to connect to the other client.

        attempts: int = 6
        - how many attempts Lynxy should make to connect to the other client.

        connection_bias: Constants.ConnectionBias = Constants.ConnectionBias.NONE
        - two Lynxy clients will typically find out an order to establish a connection in. They do this by doing a first-second order, where one follows
          the leader. However, if you want one client to always go first, then you can use a connection bias to force that. Do note that the other client does
          not know your connection bias, and will have to be programmed to accomodate the opposite bias. For example,

          >>> # client 1 has a bias of first
          >>> # client 2 has to have a bias of second
          >>> # or vice versa
        '''
        self._comm.TCP_connect(
            target_ip = target[0], 
            target_port = target[1], 
            timeout = timeout, 
            attempts = attempts,
            connection_bias = connection_bias
            )
        if start_recv: self.recv()
        return None


    # this function closes connections
    def close(self, force: bool = False) -> None: 
        '''
        When called, Lynxy will wait until all data is sent and all data is recieved before closing.
        However, if you set force to True, Lynxy will immediately terminate the connection. This can result in 
        data being lost.
        '''
        self._comm.close_connection(force)
        return None


    # this sends data
    def send(self, data: any, ignore_errors: bool = False, lock_timeout: float = 10.0) -> None: 
        '''
        When called, Lynxy will send any data you input to the other machine, encrypting it.

        data: any
        - the data you intend to send, can be anything.

        ignore_errors: bool = False
        - whether Lynxy should ignore errors and silently return or not. The following errors can be raised if set to False:
          
          >>> Lynxy.Exceptions.EmptyDataError() # if sent data is empty
          >>> lynxy.Exceptions.ClientNotConnected() # if client is not connected to other machine
          >>> Lynxy.Exceptions.SendingTimeoutError() # if client can not send after waiting for lock to be released

          All of these errors are ignored if ignore_errors is set to True.

        lock_timeout: float = 10.0
        - if data is being sent and you try to send more data, Lynxy will wait until the current message is done sending. You can set a timeout that,
          once Lynxy waits up to the timeout amount, a SendingTimeoutError will be raised.

        '''
        return self._comm.send(data, ignore_errors, lock_timeout)


    # this starts recieving data
    def recv(self) -> None:
        '''
        By default, calling the connect function will call this function, which starts recieving data
        in a background thread. However, if you want to control when you start this, then you can set start_recv in connect to False,
        and call this function when your ready.

        >>> # do this
        >>> client.connect()
        >>> # or
        >>> client.connect(start_recv = False)
        >>> client.recv() # when you're ready to start recieving
        '''
        self._comm.start_recv()
        return None

    
    # this function sets up decorators for events,
    # basically making integration with comm easier
    def event(self, eventType: Constants.Event):
        '''
        Lynxy has different events that they trigger when things happen. For example, every time a message is recieved, Lynxy will
        trigger the message event. Events can be set up as a decorators, and each event passes in different data into your function.
        Below is an example with the ON_MESSAGE event.

        >>> @client.event(lynxy.Constants.Event.ON_MESSAGE)
        >>> def woo_hoo_a_function(message: lynxy.Pool.Message):
        >>>     print(message.content)
        '''
        # wrapper function that is returned,
        # i am not quite sure how this works but it wraps around
        # the inputted function?
        def wrapper(func):
            # make a new entry for this event if it doesn't exist
            # this function will be ran everytime the event is triggered
            if eventType not in self._comm.eventRegistry.keys(): 
                self._comm.eventRegistry[eventType] = [func]
            # append function
            else: self._comm.eventRegistry[eventType].append(func)
        return wrapper


from src.lynxy import Lynxy
from rich import print
import random
import datetime

inst = Lynxy()
def recv(data):
    unpad = inst._comm.parser.removePadding(data)
    for dat in unpad:
        dec = inst._comm.sec.RSA_decrypt(dat)
        print('dec:', dec)
while True:
    current = datetime.datetime.strftime(datetime.datetime.now(), "%d/%m/%Y, %H:%M:%S")
    encrypted = inst._comm.sec.RSA_encrypt(current, True)
    padded = inst._comm.parser.addPadding(encrypted)
    # print('original:', padded)
    while padded:
        randNum = random.randint(0, len(padded))
        chars = padded[:randNum]
        padded = padded.removeprefix(chars)
        # print('current:', padded)
        # print('selected chars:', chars)
        recv(chars)


'''
This is the lynxy server module. To find documentation, go to the github~!
- Github: https://github.com/SketchedDoughnut//lynxy
'''

# extending main lynx file
from .lynx_server import *

# def help(open_tab: bool = False) -> str:
#     '''
#     A function that returns a link to the Github page
#     '''
#     if open_tab == True:
#         import webbrowser
#         webbrowser.open_new_tab('https://github.com/SketchedDoughnut/lynxy')
#     else:
#         return 'https://github.com/SketchedDoughnut/lynxy'
    
# def license(open_tab: bool = False) -> str:
#     '''
#     A function that returns a link to the license this project is under
#     '''
#     if open_tab == True:
#         import webbrowser
#         webbrowser.open_new_tab('https://github.com/SketchedDoughnut/lynxy/blob/master/LICENSE')
#     else:
#         return 'https://github.com/SketchedDoughnut/lynxy/blob/master/LICENSE'
    
# def credits() -> str:
#     '''
#     A function that returns a string containing the credits for this project
#     '''
#     return 'This project is developed and maintained by SketchedDoughnut'



# TODO
# temporary recieving function
def _recv(self) -> None:
    while True:
        # recieve how many bytes message is
        recievedNetworkOrder = self.TCP_client.recv(1024)
        if recievedNetworkOrder is None: continue # if empty ("b''")
        unpickledNetworkByteOrder = pickle.loads(recievedNetworkOrder)
        targetByteCount = socket.ntohl(unpickledNetworkByteOrder)
        # recieve byteCount amount of bytes of data
        # we load leftover to continue where it left off
        recievedData = self.parser.carry

        print('starting with:', recievedData)

        while True:
            recievedData += self.TCP_client.recv(targetByteCount)
            # ensure full data recieved
            intData = int.from_bytes(recievedData)
            recievedByteCount = intData.bit_length()

            print(f'{recievedByteCount}, {targetByteCount}, {recievedData}')

            if recievedByteCount >= targetByteCount: break
        # remove padding
        unpaddedData = self.parser.removePadding(recievedData)
        # decrypt each part
        for indivData in unpaddedData:
            print(f'decrypting ({unpaddedData.index(indivData)}):', indivData)
            decryptedData = self.sec.RSA_decrypt(indivData)
            print(f'recv ({unpaddedData.index(indivData)}):', decryptedData)
            self._trigger(Constants.Event.ON_MESSAGE, decryptedData)


# TODO
# this function sends data to the other machine
def _send(self, data: any, ignore_errors: bool = False) -> None:
    # raise error message if data is empty
    # and raise is toggled, otherwise return
    raiseError = False
    if len(data) == 0: raiseError = True
    if data is None: raiseError = True
    if not ignore_errors and raiseError: raise Exceptions.EmptyDataError()
    if ignore_errors and raiseError: return
    # find how many bytes encrypted data is
    encryptedData = self.sec.RSA_encrypt(data) # encrypt data
    paddedData = self.parser.addPadding(encryptedData) # pad data
    intData = int.from_bytes(paddedData) # get int of encoded data
    byteCount = intData.bit_length() # how many bits it takes to represent our int
    networkByteOrder = socket.htonl(byteCount) # convert to network (universal) order
    self.TCP_client.sendall(pickle.dumps(networkByteOrder)) # send length
    self.TCP_client.sendall(paddedData) # send actual data
    return


from src import lynxy
import pygame

# initialize pygame stuff
pygame.init()
window = pygame.display.set_mode((1000, 1000))
ONLINE = False

# determine who is first
first = input('1 for first, 2 for second: ') == '1'
if first: print('your color is red')
if not first: print('your color is blue')

# set up host info
host_rect_coord = (10, 500)
host_color = (255, 0, 0) if first else (0, 255, 0)

# set up opponent info
opp_rect_coord = (940, 500)
opp_color = (0, 255, 0) if first else (255, 0, 0)

# connect lynxy client
client = lynxy.Lynxy(bind=True)
print('host is:', client.get_host())
if ONLINE:
    target_ip = input('input target IP: ')
    target_port = int(input('input target port: '))
    client.connect([target_ip, target_port])
    print('connected to target')
else: print('running local')

# set up event for recieving opponent coord
@client.event(lynxy.Constants.Event.ON_MESSAGE)
def message(data: lynxy.Pool.Message):
    global opp_rect_coord
    opp_rect_coord = data.content

# main game loop
running = True
while running:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            running = False
    pressed = pygame.key.get_pressed()
    if pressed[pygame.K_w]:
        host_rect_coord = (host_rect_coord[0], host_rect_coord[1] - 1)
    elif pressed[pygame.K_s]:
        host_rect_coord = (host_rect_coord[0], host_rect_coord[1] + 1)
    window.fill((0, 0, 0))
    pygame.draw.rect(window, host_color, pygame.Rect(host_rect_coord[0], host_rect_coord[1], 50, 150))
    pygame.draw.rect(window, opp_color, pygame.Rect(opp_rect_coord[0], opp_rect_coord[1], 50, 150))
    pygame.display.update()
    if ONLINE: client.send(host_rect_coord)

pygame.quit()


from src import lynxy
import time

# set up lynxy client
client = lynxy.Lynxy(bind=True)
client.set_connection(lynxy.Constants.ConnectionType.ERROR)
print(client.get_host())

out = r'recieved.mp4'

target = ['192.168.68.129', 56774]
client.connect(target)

@client.event(lynxy.Constants.Event.ON_MESSAGE)
def on_message(data: lynxy.Pool.Message):
    f = open(out, 'wb')
    f.write(data.content)
    f.close()

time.sleep(1)

client.close()


# other imports
import pygame
import random
import time
import threading

# lynxy imports
from src import lynxy as l, lynxy_server as ls

screen_width = 500
screen_height = 500
screen_dimensions = (screen_width, screen_height)

pygame.init()
window = pygame.display.set_mode(screen_dimensions)

def gen_square():
    width = 50
    height = 50
    x = random.randint(0, screen_width - width)
    y = random.randint(0, screen_height - height)
    rect = pygame.Rect(x, y, width, height)
    return rect

# ONLY FOR THE SERVER MACHINE
def start_server() -> tuple:
    ip, port, token = ls.start_server()
    return ip, port, token

def start_connection(ip):
    success = l.start_client(ip)
    print(f'Connection status: {success}')



def send_rect():
    while True:
        time.sleep(1) 
        l.send_msg(rect1, rm)
        # print('sent:', rect)

def gen_new():
    global rect1
    while True:
        time.sleep(5)
        rect1 = ((255, 0, 0), gen_square())

def recieve_handler():
    while True:
        global rect2
        try:
            # print('setting rect2 to -1 of message queue')
            rect2 = l.message_queue[-1]
            # print('rect 2 set to:', rect2)
        except Exception as e:
            # print('set rect error:', e)
            pass



###########

# Code for the SENDING

# initialize rect
rect1 = ((255, 0, 0), gen_square()) # this is the local square
rect2 = ((0, 255, 0), gen_square()) # this is the square from the other client
# start server
ip, port, token = ls.start_server()
# start client
# input_ip = input('-> ')
start_connection(ip)
# start_connection(input_ip)
# go into listener mode
# l.send_msg('listener')  
rm = False
l.start_client_listener()
# start threads to send data
threading.Thread(target=lambda:send_rect()).start()
threading.Thread(target=lambda:gen_new()).start()
# start recieving thread
threading.Thread(target=lambda:recieve_handler()).start()
# start game loop
running = True
while running:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            running = False
    window.fill((0, 0, 0))
    pygame.draw.rect(window, rect1[0], rect1[1])
    try:
        # print('second rect drawing to:', rect2)
        pygame.draw.rect(window, rect2[0], rect2[1])
    except Exception as e:
        # print('draw error:', e)
        pass
    pygame.display.update()
pygame.quit()
ls.freeze_server()
l.shutdown_client()


################################################################## imports

# included
import socketserver
import socket
import threading
import time
import random
import pickle

# external
from cryptography.fernet import Fernet
import rsa

# file imports
# this below import is only visible because we want them to know response codes
# if they can see this, then they can see what each response code is
# hopefully leads to better understanding
# it will flood the dropdown menu of options (when they do lynxy_server.x)
from .responses import *








################################################################## variable assignment

# all valid ports it will attempt to connect to
# these ports are what the server binds to
# at least one of these ports should be the same client-side
_valid_ports = [
    11111,
    12111,
    11211,
    11121,
    11112,
    22111,
    12211,
    11221,
    11122,
    22222
]

# constants / toggles dictionary
# manages anything that can be changed 
LIMIT_USERNAME = 'limit usernames'
OVERWRITE_USERNAMES = 'ovewrite usernames'
CLEAR_DEAD_USERNAMES = 'clear dead usernames'
ENCRYPT_STORED_CLIENT_DATA = 'encrypt stored client data'
DEFAULT_PORTS = 'default ports'
OVERWRITE_IP = 'overwrite IP'
DO_PRINT = 'do print'
ALLOW_REMOTE_ADMIN = 'allow remote admin'
_toggles = {
    LIMIT_USERNAME: True,
    DO_PRINT: True,
    OVERWRITE_USERNAMES: False,
    CLEAR_DEAD_USERNAMES: True,
    ENCRYPT_STORED_CLIENT_DATA: True,
    ALLOW_REMOTE_ADMIN: False,
    DEFAULT_PORTS: _valid_ports,
    OVERWRITE_IP: None
}

# a dictionary / constants for flags
# these get changed depending on the states of the described things
_CONNECTED = 'connected'
_SHUTDOWN = 'shutdown'
_KILL_ALL = 'kill all'
_flags = {
    _CONNECTED: False,
    _SHUTDOWN: False,
    _KILL_ALL: False
}

# dictionary for storing client data
# this includes their port and ip, associated to their username
# this data is encrypted if _toggles[ENCRYPT_STORED_CLIENT_DATA] = True
_client_dict = {
    'default': 0
}

# where listener clients are stored
# these are clients who get game data distributed to them
_listener_list = []
_data_queue = []

# _HOST and _PORT info for connections
# This is the main connecting info for the server
_HOST = socket.gethostbyname(socket.gethostname())
_PORT = _valid_ports[0]

# server obj for shutting down
_server = 0

# starting thread object
_starting_thread = 0

# server token
# this is used to admin the current user and access the server
_token = 'x'

# main encryption tool for doing symmetrical encryption
_encryption_tool = Fernet(Fernet.generate_key())






################################################################## safety features 
# this includes stuff regarding encryption and whatnot

# generates auth token for remote control of server
def _gen_auth_token() -> str:
    '''
    This function generates an auth token that allows a client user
    access to the server, as an admin user.
    '''
    lower_alpha = 'abcdefghijklmnopqrstuvwxyz'
    upper_alpha = lower_alpha.upper()
    letter_list = [lower_alpha, upper_alpha]
    token = ''
    for i in range(6): # length of your sign-in token
        letter_or_num = random.randint(0, 1)
        if letter_or_num == 0:
            letter_type = random.randint(0, 1)
            letter_range = letter_list[letter_type]
            letter_index = random.randint(0, len(letter_range) - 1)
            token += letter_range[letter_index]
        elif letter_or_num == 1:
            token += str(random.randint(0, 9))
    return token

# generates the public and private key for server
def _gen_access_keys() -> tuple:
    public_key, private_key = rsa.newkeys(1024)
    return public_key, private_key

# encrypts data for storage
def _encrypt_data(data: bytes) -> bytes:
    # returns encrypted data
    return _encryption_tool.encrypt(data)

# decrypts data from storage
def _decrypt_data(data: bytes) -> bytes:
    # returns decrypted data
    return _encryption_tool.decrypt(data)

# uses clients public key to encrypt data
def _encrypt_public(cpk, data) -> bytes:
    pickled_data = pickle.dumps(data)
    return rsa.encrypt(pickled_data, cpk)

# uses servers private key to decrypt data
def _decrypt_private(data) -> any:
    data = rsa.decrypt(data, _server_private_key)
    unpickle_data = pickle.loads(data)
    return unpickle_data

















# function data
# _do_custom_function = False
# _init_custom_function = 0
# _custom_function = 0





















## OVERRIDE FUNCTIONS
# override ports
def override_ports(ports: list) -> None:
    ''' 
    Overrides what ports the server will attempt to connect to
    '''
    global _valid_ports, _ov_ports
    _ov_ports = ports

# override ip
def override_ip(ip_in: str) -> None:
    '''
    Overrides what IP the server will attempt to connect to
    '''
    global _ov_ip
    _ov_ip = ip_in

def set_limit_username(choice: bool) -> None:
    global limit_username
    limit_username = choice

def set_overwrite_usernames(choice: bool) -> None:
    global overwrite_usernames
    overwrite_usernames = choice

def set_clear_dead_usernames(choice: bool) -> None:
    global clear_dead_usernames
    clear_dead_usernames = choice

def set_encrypt_client_data(choice: bool) -> None:
    global encrypt_client_data
    encrypt_client_data = choice

# disable prints
def disable_print() -> None:
    '''
    Disables the server from printing messages
    '''
    global _do_print
    _do_print = False

# enable prints
def enable_print() -> None:
    '''
    Enables the server to print messages
    '''
    global _do_print
    _do_print = True


## FEATURE FUNCTIONS
# function to handle printing
def pprint(msg: str) -> None:
    '''
    A function meant for filtering prints based on if it is enabled or disabled - This is meant for internal use
    '''
    if _do_print:
        print(msg)
    else:
        pass

def _fancy_send(client: socket.socket, cpk, data) -> None:
    '''
    A function meant for fancy sending
    '''
    # print('SEND INVOKED)')
    # pickled_data = pickle.dumps(data)
    encrypted_data = _encrypt_public(cpk, data)
    client.sendall(encrypted_data)

def _fancy_recieve(client: socket.socket) -> any:
    '''
    A function meant for fancy recieving
    '''
    incoming = client.recv(1024)
    decoded = _decrypt_private(incoming)
    # print('FANCY RECIEVE: decode:', decoded)
    return decoded


def _log_user_data(key: str, data: tuple) -> None:
    global _client_dict
    if encrypt_client_data == True: # if server set to encrypt data
        # print(encrypt_client_data)
        # string_data = str(data) # convert data to string
        pickled_data = pickle.dumps(data)
        # encoded_data = string_data.encode() # encode string to bytes
        encrypted_data = _encrypt_data(pickled_data) # encrypt bytes into bytes
        data = encrypted_data
    _client_dict[key] = data # log

def _get_user_data(key: str) -> tuple:
    data = _client_dict[key] # get current data 
    if encrypt_client_data == True: # if server set to encrypt data
        decrypted_data = _decrypt_data(data) # data is already bytes, directly decrypt
        # string_data = decrypted_data.decode() # convert bytes to string
        string_data = pickle.loads(decrypted_data)
        data = string_data # set data to tuple
    return data # return for send

def _remove_dead(username: str):
    if clear_dead_usernames:
        try:
            del _client_dict[username]
        except:
            pass 

# from typing import Callable
# def load_function(loop_function: Callable) -> None:
#     global _do_custom_function, _custom_function
#     _do_custom_function = True
#     _custom_function = loop_function


# function to display current data
def get_data() -> dict:
    '''
    Returns data about the current server in the form of a dictionary
    '''
    return {
        'server info': {
            'is_alive': _connected,
            'ip': _HOST,
            'port': _PORT,
            'token': _token
        },
        'client info': _client_dict,
        'listener info': _listener_list,
        'message_queue': _data_queue
    }

# function for shutting down the server
def freeze_server(do_print: bool = True) -> str:
    '''
    A function to shut down the server: returns a status code.
    '''
    global _server, _kill_all
    try:
        _server.shutdown()
        _kill_all = True
        if do_print == True:
            pprint('[SERVER SHUTDOWN] Shutting down server...')
        return OPERATION_SUCCESS
    except:
        return OPERATION_FAIL
    
# function to poll shutdown var, if it is enabled then shutdown
def _poll_shutdown() -> None:
    global _kill_all
    while True:
        if _shutdown == True:
            pprint('[SERVER SHUTDOWN] request to shutdown detected, shutting down server...')
            freeze_server(False)
            pprint('[SERVER SHUTDOWN] Enabling _kill_all...')
            _kill_all = True
            pprint('[SERVER SHUTDOWN] Server shut down, exiting...')
            break
        time.sleep(5)
    exit()

## SAFETY FUNCTIONS


## FUNCTIONS FOR DISTRIBUTION SERVER
def _loopback_input(client: socket.socket, cpk) -> None:
    # become brick
    time.sleep(0.025)
    while True:
        try:
            # msg = client.recv(1024).decode()
            msg = client.recv(1024)
            decoded = _decrypt_private(msg)
            # client.sendall(OPERATION_SUCCESS)
            # if isinstance(decoded, str) == True and decoded.isspace() == True:
            #     continue
            # elif not decoded:
            #     continue
            _data_queue.append([client, decoded])
            # client.sendall(OPERATION_SUCCESS)
            # _fancy_send(client, cpk, OPERATION_SUCCESS)
        except:
            # packet loss
            pass

def _distributor() -> None:
    global _data_queue
    while True:
        for message_data in _data_queue:
            ignore_client = message_data[0]
            message = message_data[1]
            for client in _listener_list:
                client_key = client[1]
                active_client = client[0]
                if active_client != ignore_client: 
                # if active_client == ignore_client: # TESTER STATEMENT
                    encoded = _encrypt_public(client_key, message)
                    try:
                        # new message segments
                        segmented = f'{encoded}~E'
                        encoded_segment = segmented.encode()
                        # active_client.sendall(encoded)
                        active_client.sendall(encoded_segment)
                    except:
                        # assume the client cant be commmunicated with, remove from ppl to send to
                        _listener_list.remove(client)
            _data_queue.remove(message_data)
            # time.sleep(0.0025) # who needs a delay anyways?

# MAIN CLASS
class _myTCPserver(socketserver.BaseRequestHandler):
    def handle(self) -> None:
        global _client_dict, _shutdown

        is_listener = False
        _verified = False
        # local_data_queue = []
        # local_data_save = []
        saved_username = ''











        pprint('[SERVER HANDSHAKE] Handshaking with client...')
        # send public key of server to client
        pickled_server_public_key = pickle.dumps(_server_public_key)
        self.request.sendall(pickled_server_public_key)

        # recieve clients public key
        pickled_client_public_key = self.request.recv(1024)
        client_public_key = pickle.loads(pickled_client_public_key)

        # print('server priv key:', _server_private_key)
        # print('server pub key:', _server_public_key)
        # print('server client pub key:', client_public_key)















        
        while True:
            # establish client address
            addr = self.client_address[0]
            addr = self.client_address[1]

            # kill client communication if is true (will kill before msg)
            if _kill_all == True:
                # self.request.sendall('the server has been commanded to kill all client instances'.encode())
                # self.request.sendall(KILL_ALL)
                _fancy_send(self.request, client_public_key, KILL_ALL)
                pprint(f'[{addr}] Killing this instance, due to _kill_all being True...')
                _remove_dead(saved_username)
                break

            # format incoming message
            if not is_listener:
                try:
                    # msg = self.request.recv(1024)
                    # msg = msg.decode()
                    msg = _fancy_recieve(self.request)
                    if not msg:
                        _fancy_send(self.request, client_public_key, INVALID_MESSAGE)
                        continue
                    split_msg = msg.split()
                    prefix = split_msg[0]
                    split_msg.remove(prefix)
                    joined_msg = "".join(split_msg)
                except:
                    try:
                        # self.request.sendall(INVALID_MESSAGE) # try to send message telling them what they gave is invalid
                        _fancy_send(self.request, client_public_key, INVALID_MESSAGE)
                        continue
                    except Exception as e:
                        pprint(f'[{addr}] - crash - ending this instance')
                        _remove_dead(saved_username)
                        pprint('----------------------------------------------')
                        break
            
            if is_listener:
                # if len(_data_queue) > len(local_data_queue): # if there is a difference in the lists
                #     dif = len(_data_queue) - len(local_data_queue)
                #     for i in range(dif): # for amount of difference
                #         local_data_queue.append(_data_queue[-1]) # add all different objects to local data queue
                #         print('adding dif:', _data_queue[-1])
                #     print('local queue:', local_data_queue)
                #     # exit()
                #     for item in local_data_queue: # for each msg, encode 
                #         item = item.encode()
                #         for client in _listener_list: # for each client, send encoded message
                #             print('sending to client:', client)
                #             client.sendall(item)
                #         local_data_save.append(item.decode()) # save to local data save
                #         local_data_queue.pop(0) # remove most recent
                #         print('data save:', local_data_save)
                #         print('data queue:', local_data_queue)
                #     local_data_queue = []
                continue

            # if _do_custom_function:
            #     result = _custom_function(msg, self.client_address)
            #     if result == 'continue':
            #         continue
            #     elif result == 'break':
            #         _remove_dead(saved_username)
            #         break
            #     elif result == 'exit':
            #         _remove_dead(saved_username)
            #     else:
            #         pass
            # if prefix is username, log their username and their device info (ip, port) associated with it
            if prefix == 'username':
                # print('is username prfix')
                if joined_msg: # if not empty
                    # print('isnt empty')
                    # if they have not already inputted a username and the username limit is not on,
                    if limit_username == True: # if username limit is enabled
                        # print('do limit username')
                        if saved_username: # if they have already submtited a username,
                            # print('do save username')
                            # self.request.sendall(ALREADY_REGISTERED_USERNAME) # tell them they have already registered a username
                            # print('already registered')
                            _fancy_send(self.request, client_public_key, ALREADY_REGISTERED_USERNAME)
                            continue # start next loop iteration
                        else: # else, they have not submitted a username
                            # print('dont save username')
                            if overwrite_usernames == True: # if they want to overwrite usernames
                                # print('overwrite usernames)')
                                pprint(f'[{addr}] {prefix} - logging {self.client_address} to {joined_msg}')
                                # _client_dict[joined_msg] = self.client_address # write username regardless
                                # _log_user_data(joined_msg, (self.client_address, self.request))
                                _log_user_data(joined_msg, self.client_address)
                                saved_username = joined_msg
                                # self.request.sendall(OPERATION_SUCCESS) # send back success
                                # print('op success')
                                _fancy_send(self.request, client_public_key, OPERATION_SUCCESS)
                                continue # start next loop iteration
                            else: # if they do not want to overwrite usernames
                                # print('dont overwrite usernames')
                                try: # see if the entry exists
                                    _client_dict[joined_msg] # try to access entry
                                    # self.request.sendall(USERNAME_EXISTS) # if entry exists, no error, returns exists
                                    # print('username exists')
                                    _fancy_send(self.request, client_public_key, USERNAME_EXISTS)
                                    continue # start next loop iteration
                                except KeyError: # fails since no key exists
                                    # _client_dict[joined_msg] = self.client_address # set username since it does not exist
                                    _log_user_data(joined_msg, self.client_address)
                                    pprint(f'[{addr}] {prefix} - logging {self.client_address} to {joined_msg}')
                                    saved_username = joined_msg
                                    # self.request.sendall(OPERATION_SUCCESS)
                                    # print('op success')
                                    _fancy_send(self.request, client_public_key, OPERATION_SUCCESS)
                    else:
                        # print('dont limit usernames')
                        if overwrite_usernames == True:
                            # print('overwrite usernames')
                            pprint(f'[{addr}] {prefix} - logging {self.client_address} to {joined_msg}')
                            # _client_dict[joined_msg] = self.client_address # write username regardless
                            _log_user_data(joined_msg, self.client_address)
                            saved_username = joined_msg
                            # self.request.sendall(OPERATION_SUCCESS) # send back success
                            # print('op success')
                            _fancy_send(self.request, client_public_key, OPERATION_SUCCESS)
                            continue # start next loop iteration
                        else:
                            # print('dont overwrite usernames')
                            try: # see if the entry exists
                                _client_dict[joined_msg] # try to access entry
                                # self.request.sendall(USERNAME_EXISTS) # if entry exists, no error, returns exists
                                _fancy_send(self.request, client_public_key, USERNAME_EXISTS)
                                continue # start next loop iteration
                            except KeyError: # fails since no key exists
                                # _client_dict[joined_msg] = self.client_address # set username since it does not exist
                                pprint(f'[{addr}] {prefix} - logging {self.client_address} to {joined_msg}')
                                _log_user_data(joined_msg, self.client_address)
                                saved_username = joined_msg
                                # self.request.sendall(OPERATION_SUCCESS)
                                # print('op success')
                                _fancy_send(self.request, client_public_key, OPERATION_SUCCESS)

                        # pprint(f'[{addr}] {prefix} - logging {self.client_address} to {joined_msg}')
                        # # self.request.sendall('logged username, data'.encode())
                        # self.request.sendall(OPERATION_SUCCESS)
                else:
                    # self.request.sendall(INVALID_MESSAGE)
                    # print('invalid message')
                    _fancy_send(self.request, client_public_key, INVALID_MESSAGE)

            # if prefix is request_by_user, attempt to return the data associated with that username. If it does not exist, send back "None"
            elif prefix == 'request_by_user':
                try:
                    # self.request.sendall(str(_client_dict[joined_msg]).encode())
                    d = _get_user_data(joined_msg)
                    # self.request.sendall(d2)
                    _fancy_send(self.request, client_public_key, d)
                    pprint(f'[{addr}] {prefix} - return {joined_msg} data: {d}')
                except:
                    pprint(f'[{addr}] {prefix} - return {joined_msg} data: None')
                    # self.request.sendall('None'.encode())
                    # self.request.sendall(INVALID_USERNAME_DATA)
                    _fancy_send(self.request, client_public_key, INVALID_USERNAME_DATA)

            # if prefix is auth, check if token is matching, then allow user to use dev features
            elif prefix == 'auth':
                if _toggles[ALLOW_REMOTE_ADMIN] == False:
                    _fancy_send(self.request, client_public_key, FEATURE_DISABLED)
                    continue
                if joined_msg == _token:
                    _verified = True
                    # self.request.sendall('client session authorized'.encode())
                    # self.request.sendall(OPERATION_SUCCESS)
                    _fancy_send(self.request, client_public_key, OPERATION_SUCCESS)
                    pprint(f'[{addr}] {prefix} - authed client')
                else:
                    # self.request.sendall('invalid auth token'.encode())
                    # self.request.sendall(INVALID_AUTH_TOKEN)
                    _fancy_send(self.request, client_public_key, INVALID_AUTH_TOKEN)

            elif msg == 'help':
                msg = 'Commands: username, request_by_user, auth, help, clear_client (auth only), freeze_server (auth only)'
                # encoded = msg.encode()
                # self.request.sendall(encoded)
                _fancy_send(self.request, client_public_key, msg)
                
            # if msg is listener, add their socket to the listening list and ignore any more messages from them
            elif msg == 'listener':
                if not is_listener:
                    _listener_list.append([self.request, client_public_key])
                    is_listener = True
                    # threading.Thread(target=lambda:_loopback_input(self.request, client_public_key)).start()
                    # self.request.sendall(OPERATION_SUCCESS)
                    # time.sleep(0.025)

                    # lets create a brick
                    _fancy_send(self.request, client_public_key, OPERATION_SUCCESS)
                    pprint(f'[{addr}] {msg} - subscribing to listener')
                    _loopback_input(self.request, client_public_key)


            # if msg is clear_client, check if this client is authorized and then clear the client_dict
            elif msg == 'clear_client':
                if _verified == True:
                    _client_dict = {
                        'default': 0
                    }
                    # self.request.sendall('cleared client dictionary')
                    # self.request.sendall(OPERATION_SUCCESS)
                    _fancy_send(self.request, client_public_key, OPERATION_SUCCESS)
                    pprint(f'[{addr}] {msg} - clearing client_dict')
                else:
                    # self.request.sendall('user not authorized'.encode())
                    # self.request.sendall(USER_NOT_AUTHORIZED)
                    _fancy_send(self.request, client_public_key, USER_NOT_AUTHORIZED)


            # if msg is freeze_server, check if this client is authorized and then raise the flag to shutdown srever
            elif msg == 'freeze_server':
                if _verified == True:
                    _shutdown = True
                    # self.request.sendall('shutdown of server requested, raising flag'.encode())
                    # self.request.sendall(OPERATION_SUCCESS)
                    _fancy_send(self.request, client_public_key, OPERATION_SUCCESS)
                    pprint(f'[{addr}] {msg} - shutdown of server requested, raising flag')
                else:
                    # self.request.sendall('user not authorized'.encode())
                    # self.request.sendall(USER_NOT_AUTHORIZED)
                    _fancy_send(self.request, client_public_key, USER_NOT_AUTHORIZED)

            # if msg is end_session, end the current session the server and the client have
            elif msg == 'end_session':
                # self.request.sendall('ending'.encode())
                # self.request.sendall(END_SESSION)
                _fancy_send(self.request, client_public_key, END_SESSION)
                pprint(f'[{addr}] {msg} - ending this instance')
                # print("CHECKING CLEAR")
                _remove_dead(saved_username)
                pprint('----------------------------------------------')
                break
            
            # ignore their message if otherwise
            else:
                # self.request.sendall(msg.upper().encode())  # Send response back to the client
                # self.request.sendall('invalid command'.encode())
                # self.request.sendall(INVALID_COMMAND)
                _fancy_send(self.request, client_public_key, INVALID_COMMAND)
                pass











# FINAL VAR DEFINITIONS

# private / public key setup
_server_public_key, _server_private_key = _gen_access_keys()








# STARTING CODE

# main function for starting, does not use a thread and will block code
def no_thread_start_server(is_threaded: bool = False) -> None:
    '''
    If you want to start the server without it running in a thread, you can call this function. However, this will block your code until the server goes offline.
    This won't happen unless it crashes, or you remotely raise the shutdown flag (refer to the server setup page of documentation on Github)
    '''
    global _HOST, _PORT, _valid_ports, _connected, _server, _token
    ## apply overrides
    # override ip
    if _ov_ip:
        _HOST = _ov_ip
        pprint(f'[OVERRIDE] IP overrided to: {_HOST}')
    if len(_ov_ports) > 0:
        _valid_ports = _ov_ports
        _PORT = _valid_ports[0]
        pprint(f'[OVERRIDE] Valid ports overrided to: {_valid_ports}')

    # pre-loop variables
    _connected = False
    
    # generate unique session token for remote controlling the server
    _token = _gen_auth_token()

    # loop, trying to find a free port
    for port in _valid_ports:

        try:
            pprint(f'[PORT CYCLE] Server trying port: {port}')
            with socketserver.ThreadingTCPServer((_HOST, port), _myTCPserver) as _server:
                pprint(f'[PORT CYCLE] Server found port for startup: {port}')
                # start server shutdown poll
                threading.Thread(target=lambda:_poll_shutdown()).start()    # , daemon=True).start()
                pprint('[SERVER] Started scan for shutdown requests')
                # start distributor thread
                threading.Thread(target=lambda:_distributor()).start()
                pprint('[SERVER] Started distributor thread')
                if is_threaded: 
                    pprint(f'[SERVER] Server IP: {_HOST}')
                    pprint(f'[SERVER] Control token: {_token}')
                else: 
                    print(f'[SERVER] Server IP: {_HOST}')
                    print(f'[SERVER] Session token: {_token}')
                pprint('[SERVER] Server is ready for communication~!')
                if is_threaded:
                    pprint('----------------------------------------------')
                else: 
                    print('----------------------------------------------')
                _connected = True
                _PORT = port
                _server.serve_forever()
                break
        except:
            try:
                pprint(f'[PORT CYCLE] Server port cycling: {port} -> {_valid_ports[_valid_ports.index(port) + 1]}')
            except IndexError:
                port = _valid_ports[0]
                pprint(f'[PORT CYCLE - RESET 2] Server resetting port to: {port}')
    
    if _connected == False:
        pprint('[PORT CYCLE - ERROR 0] Server failed to find an open valid port, exiting')
        exit()
    else:
        pprint('[PORT CYCLE - ERROR 1] It is assumed server has been shutdown, ignoring error')





# starts the server via a thread, to let the code calling this function continue running instead of blocking
def start_server() -> tuple:
    global _starting_thread
    '''
    Starts the server in a thread, which means this will not block the rest of your code if you have more things done after this function is called. 
    This function also returns the IP that the server is on, the port the server is on, and the authorization token in a tuple.
    '''
    _starting_thread = threading.Thread(target=lambda:no_thread_start_server(True)) #, daemon=True)
    _starting_thread.start()
    time.sleep(0.25) # this is to not get false information if they request data later on 
    return _HOST, _PORT, _token


# import lynxy as l
# import lynxy_server as ls

from src import lynxy as l
from src import lynxy_server as ls

def run_server():
    # ls.set_encrypt_client_data(False)
    ip, port, token = ls.start_server()
    return (ip, port, token)

def run_client(ip):
    respond_mode = True
    l.enable_print()
    # ip = input('-> ')
    print('state of connect:', l.start_client(ip))
    while True:
        msg = input('-> ')
        # msg = 't'
        if msg == 'break':
            break
        elif msg == 'token':
            l.send_msg(f'auth {token}')
            continue
        elif msg == 'listener':
            respond_mode = False
        l.send_msg(msg, respond_mode)
    l.shutdown_client()


# running
ip, port, token = run_server()
run_client(ip)


################################################################## imports

# included
import socket
import pickle
import threading

# external
import rsa








################################################################## variable assignment

# default ports client will try and connect to
_valid_ports = [
    11111,
    12111,
    11211,
    11121,
    11112,
    22111,
    12211,
    11221,
    11122,
    22222
]

# constants / toggles dictionary
# manages anything that can be changed 
DO_PRINT = 'print'
DEFAULT_PORTS = 'port_override'
_toggles = {
    DO_PRINT: False,
    DEFAULT_PORTS: _valid_ports
}

# establish the main socket variables
# such as server ip, server port, main_client communicating with server
_HOST = '' # server ip
_PORT = _valid_ports[0] # server port
# _HOST, _PORT = '', _valid_ports[0] # server ip, port ip
_main_client = socket.socket(socket.AF_INET, socket.SOCK_STREAM) # main client for communication with server

# status variables
# these help know what has succeeded and whatnot
_connected = False # connected to server

# safety vars
# these are used for encryption and general safety
_server_public_key = 0
_client_private_key = 0
_client_public_key = 0

# listener features for when data is distributed to clients from the other end
# this is publicly accessible so users can get data out of it
data_queue = []








################################################################## optional functions for overriding / toggles

# override ports
# used to change the ports the client tries to connect to
# def override_ports(ports: list) -> None:
#     ''' 
#     Overrides what ports the client will attempt to connect to. \n
#     ARGS
#         - ports: list
#         Each port should be an integer, and the server and the client should have at least one port in common.
#     '''
#     # global _ov_ports
#     # _ov_ports = ports
#     global _toggles
#     _toggles[DEFAULT_PORTS] = ports

# def disable_print() -> None:
#     '''
#     Disables the client from printing messages
#     '''
#     global _do_print
#     _do_print = False

# # enable prints
# def enable_print() -> None:
#     '''
#     Enables the client to print messages
#     '''
#     global _do_print
#     _do_print = True

# toggles things that are customizable
# this includes the ports to connect to, and the option to print to console
# identified as DEFAULT_PORTS and DO_PRINT
def toggle(toggle, state) -> None:
    '''
    Edits the value associated with the input "toggle". \n
    # Parameters
    **toggle** \n
    This should be set to the constants provided by lynxy, to signify what you want to change. They are listed below!
        - DO_PRINT
            - information: 
                - this toggles whether lynxy prints message to console or not. This is disabled by default.
            - values: True, False
            - example: 
                - toggle(DO_PRINT, True)
                - toggle(DO_PRINT, False)
        - DEFAULT_PORTS
            - information:
                - this regards the ports that the client will attempt to connect to on the servers IP.
            - values: list of integer ports
            - example: 
                - toggle(DEFAULT_PORTS, [12345, 23456, 34567])
            - notes: The client and the server HAVE to have at least one port in common! This is very important, 
                     because if they do not have a common port then the two can not connect. \n
    **state** \n
    State should be the corresponding value for what you entered for toggle, meaning that if for example you do `DO_PRINT`, 
    then state should either be `True` or `False`. More specific cases for states are shown above.
    '''
    global _toggles, _valid_ports
    _toggles[toggle] = state
    _valid_ports = _toggles[DEFAULT_PORTS] # update valid ports manually (idk why?)








################################################################## other general use functions (internal)

# function to handle printing if its enabled
def pprint(msg: str) -> None:
    '''
    A function meant for filtering prints based on if it is enabled or disabled - this is meant for internal use.
    # Parameters
    **msg** \n
    the information to print
    '''
    if _toggles[DO_PRINT]:
        print(msg)








################################################################## other general use functions (external)

# function to display current data
def get_data() -> dict:
    '''
    This is the main function used for acquiring data used by lynxy. The data it returns is as follows:
    - server ip
    - server port
    - ports to connect to
    - main client socket.socket object
    - boolean representing connected or not
    - toggles, including whether to print or not and the ports to connect to
    - security data, including the clients public and private key, and the servers public key
    # Returns
    - dictionary with above data
    '''
    return {
        'server info': {
            'ip': _HOST,
            'port': _PORT
        },
        'client info': {
            'default ports': _valid_ports,
            'main client': _main_client,
            'connected': _connected,
            'toggles': _toggles
        },
        # 'data': {
        #     'data_queue': data_queue
        # },
        'security': {
            'client public key': _client_public_key,
            'client private key': _client_private_key,
            'server public key': _server_public_key
        },
        'sillies': 'sillies :3'
    }








################################################################## functions for safety / security (encryption and whatnot)

# function to generate RSA public and private keys
def _gen_access_keys() -> tuple[rsa.PublicKey, rsa.PrivateKey]:
    '''
    Generates a public and private key, and returns them in a tuple like this: [public_key, private_key]
    '''
    public_key, private_key = rsa.newkeys(1024)
    return public_key, private_key


# function to do a handshake with the server, exchanging public and private keys
def _handshake(client: socket.socket) -> None:
    '''
    Does a handshake with the server, exchanging public keys.
    # Parameters
    - client: socket.socket object
    '''
    # globals
    global _client_public_key, _client_private_key # globally edit the clients public and private key
    global _server_public_key # globally edit the servers private key
    # generate keys
    pprint('[CLIENT HANDSHAKE] Handshaking with server...')
    _client_public_key, _client_private_key = _gen_access_keys() # generate public and private keys for client
    # acquire server public key
    pickled_server_public_key = client.recv(1024) # recieve pickled (pickle module turns var into bytes (var can be string, object, whatever)) public key from server
    _server_public_key = pickle.loads(pickled_server_public_key) # converts the bytes object back into its original form
    # send public key to server
    pickled_client_public_key = pickle.dumps(_client_public_key) # convert the public_key into bytes
    client.sendall(pickled_client_public_key) # send to server


# function to encrypt data when sending to server, using the servers public key
def _encrypt_public(data) -> bytes:
    '''
    A function that encrypts data using RSA, and the servers public key.
    # Parameters
    - data: any
    # Returns
    - bytes
    '''
    pickled_data = pickle.dumps(data) # converts data into bytes
    encoded_data = rsa.encrypt(pickled_data, _server_public_key) # encrypts data with servers public key
    return encoded_data


def _decrypt_private(data) -> any:
    '''
    A function that decrypts using RSA, and the clients private key
    '''
    decrypted_data = rsa.decrypt(data, _client_private_key) # decrypts data with the clients private key
    loaded_data = pickle.loads(decrypted_data) # converts bytes into data
    return loaded_data








################################################################## functions used for connecting to server (internal)

# cycles ports until a connection to the server succeeds / fails.
def _cycle_port(client: socket.socket) -> tuple[socket.socket, int, bool]:
    '''
    An internal function used to cycle through the ports in _valid_ports to try and make a connection.
    # Parameters
    - client: socket.socket object
    # Returns
    - tuple[socket.socket object of server, the port the client connected to, a boolean on whether the connection succeeded or not]
    '''
    # globals
    global _connected # indicates if the client has connected or not
    _connected = False # by default, set to False
    for port in _valid_ports: # iterate through each port
        try: # try and connect to server with current port
            pprint(f'[PORT CYCLE] Client trying port: {port}')
            client.connect((_HOST, port)) # try to connect to server ip, and the current cycled port
            pprint(f'[PORT CYCLE] Client connected to: {port}')
            pprint('----------------------------------------------')
            _connected = True # connection succeeded, set variable
            _handshake(client) # do handshake with server to exchange public keys
            return client, port, _connected # returns the client communicating with server, the port that it is connected to, and a boolean for the connection status
        except: # connection to that port failed
            try: # print that we are trying next port
                current_port_index = _valid_ports.index(port) 
                next_port = _valid_ports[current_port_index + 1]
                pprint(f'[PORT CYCLE] Client port cycling: {port} -> {next_port}')
            except IndexError: # failed, pass since the for loop will exit after this iteration is done anyways
                pass
    if _connected == False: # verify _connected is False
        pprint('[PORT CYCLE] the client can not find a open valid server port, exiting')
        return client, _PORT, _connected # return the client, _PORT as it has not changed, and a boolean for the connection status








################################################################## functions used for connecting to server / other client (external)

def start_client(connection_ip: str) -> bool:
    '''
    Starts a connection with the other end.
    # Parameters
    **connection_ip**
    the ip of the other end to connect to as a string
    # Returns
    a boolean saying whether the connection succeeded or not
    '''
    # globals
    global _main_client # client for communicating with server
    global _valid_ports # edit the valid ports for the client to connect to
    global _HOST, _PORT # servers ip, and port
    _HOST = connection_ip # set global server ip to the inputted connection_ip
    # establish connection
    _main_client = socket.socket(socket.AF_INET, socket.SOCK_STREAM) # reset the main_client socket
    _valid_ports = _toggles[DEFAULT_PORTS] # set the valid ports to the ports inside of _toggles
    _main_client, _PORT, state = _cycle_port(_main_client) # establish the connection to a port that the server is on
    return state # whether the connection succeeded or not








################################################################## functions for communicating with server / other client

# sends data to server, gets data back if recieve = True
def send_msg(data: any, recieve: bool = True) -> any:
    '''
    A general messaging function for sending data to the other end (decided by what _main_client is connected to).
    # Parameters
    **data** \n
    The information you want to send to the other end. This can be anything you want! \n
    **recieve** \n
    True by default, set this to False if you do not want send_msg() to wait for a response from the other end.
    # Returns
    data recieved from other end, if recieve = True
    '''
    encoded_data = _encrypt_public(data)
    _main_client.sendall(encoded_data)
    pprint(f"Sent:     {data}")
    if recieve:
        incoming_data = _main_client.recv(1024)
        decrypted_data = _decrypt_private(incoming_data)
        pprint(f"Received: {decrypted_data}")
        return decrypted_data


# a function for submitting username data to the server
def submit_username_data(username: str) -> str:
    '''
    Submits a username to the server, which the server will associate with your IP and port. \n
    # Parameters \n
    **username** \n
    Your username as a string (please, no spaces!) \n
    # Returns \n
    a status code as a string that gives information on what happened server-side. \n
    Please note this function is meant for use only with a lynxy_server server. Otherwise, your server must be configured 
    to handle this call.
    '''
    full_data = f'username {username}'
    encrypted_data = _encrypt_public(full_data)
    _main_client.sendall(encrypted_data)
    pprint(f"Sent:     {full_data}")
    incoming_data = _main_client.recv(1024)
    decrypted_data = _decrypt_private(incoming_data)
    pprint(f"Received: {decrypted_data}")
    return decrypted_data


# requests ip and port from server
def request_username_data(username: str) -> any:
    '''
    Requests the data associated with a username from the server. 
    # Parameters
    **username** \n
    a string of the username that you want to request the data with (Please, no spaces!) \n
    # Returns \n
    a status code as a string that gives information on what happened server-side. \n
    If the server has disabled this feature, you will instead get a status code signifying a fail. 
    This might occur if the server has disabled directly connecting to other clients. 
    If this happens, then you will have to communicate with the other clients by directing traffic through the server. \n
    Please note this function is meant for use only with a lynxy_server server. Otherwise, your server must be configured 
    to handle this call.
    '''
    full_data = f'request_by_user {username}'
    # encoded_message = message2.encode('utf-8') # added request_by_user prefix by default
    encoded_message = _encrypt_public(full_data)
    _main_client.sendall(encoded_message)
    pprint(f"Sent:     {full_data}")
    # incoming_data = full_recieve(client)
    # incoming_data = client.recv(1024).decode('utf-8')
    incoming_data = _main_client.recv(1024)
    decrypted_data = _decrypt_private(incoming_data)
    pprint(f"Received: {decrypted_data}")
    return decrypted_data

    # parse into list
    # address_str = incoming_data.strip('()')
    # ip_str, port_str = address_str.split(',')
    # ip_str = ip_str.strip().strip("'")
    # port_int = int(port_str.strip())
    # incoming_data = [ip_str, port_int]








################################################################## functions for ending connections with the other end

# function for shutting down the client
def shutdown_client() -> bool:
    '''
    A function that shuts down the clients connection with the other end.
    # Returns
    a boolean telling you whether or not the shutdown succeeded
    '''
    global _main_client
    try: send_msg('end_session')
    except: pass
    try:
        _main_client.close()
        pprint('[CLIENT SHUTDOWN] Shutting down client...')
        return True
    except: return False








################################################################## functions for starting other processes

def start_listening() -> None:
    '''
    A function that acts on the recieving end to recieve information from the other end (decided by what _main_client is connected to). 
    It adds to a data queue that can be accessed to retrieve data. The latest entry in the data queue is the newest information. 
    '''
    # "The queue gets cleared after 25 entries for efficiency and optimal storage."
    send_msg('listener')
    threading.Thread(target=lambda:_inbound_data_listener()).start() # listener which recieves data
    threading.Thread(target=lambda:_inbound_data_parser()).start() # parser which identifies individual commands
    threading.Thread(target=lambda:_parsed_data_decrypter()).start() # decrypter which decrypts individual commands, this one adds to data_queue
    # threading.Thread(target=lambda:_data_queue_cleaner()).start() # cleaner for list, saves latest val








################################################################## functions / variables that are part of the listening category

_inbound_data = []
_parsed_data = []


# this function is responsible for decrypting and finally outputting data to the message queue
def _parsed_data_decrypter():
    '''
    A function that decrypts the data put into _parsed_data_decrypter, then pipes that data into data_queue.
    '''
    global _parsed_data # so we can erase current _parsed_data after making a local copy
    while True:
        lc_parsed_data = _parsed_data # lc = local_copy
        _parsed_data = [] # reset _parsed_data as we have copied locally, there is the potential for some data loss here
        for parsed_data in lc_parsed_data:
            try:
                decoded = _decrypt_private(parsed_data)
                data_queue.append(decoded)
            except Exception as e: # this error will likely happen if the parser fails to identify and assemble a proper message
                print('NEW DECRYPTION LOST:', e)
                pass
            # we do not need to remove data as that will mess with iteration
            # and once we are done iterating this data will be erased anyways
            # lc_parsed_data.remove(parsed_data) 


# this function creates a parser which identifies the start and end of a full message, joins that, then pipes it into _parsed_data
def _inbound_data_parser():
    '''
    A function that identifies the start and end of each full data packet, pieces them together, and then pipes that into _parsed_data.
    '''
    global _inbound_data
    lc_inbound_data = _inbound_data # lc = local copy
    _inbound_data = [] # reset _inbound_data as we have copied locally, there is the potential for some data loss here
    # im not really sure what happens from here down, but i think it works so oh well- not sure how i even wrote this
    # i tried to comment as best i can
    # i also renamed some stuff
    while True:
        for data in lc_inbound_data: # iterate over each piece of packet stored
            data: bytes # type hinting
            decoded_data = data.decode() # decode data
            split_entry = [str(i) for i in decoded_data] # split it into each character
            new_packet = '' # this represents the new full packet of data
            for letter in split_entry: # iterate over each character
                if letter == '~': # first marker of the end flag
                    location = split_entry.index(letter) # second marker of end flag
                    try: # try to check next character, if fails then at end of list
                        next = split_entry[location + 1]
                        if next == 'E':
                            # new_packet = new_packet.encode()
                            new_packet = eval(new_packet) # turn back into bytes??
                            _parsed_data.append(new_packet) # pipe full packet into _parsed_data
                            new_packet = '' # reset new_packet
                            break # break out of loop, currently doesn't continue searching
                    except IndexError as e:
                        print('NEW SEGMENT LOST:', e) # lost packed because could not find second part of end marker
                else:
                    new_packet += letter # add this character to the new packet
            # we do not need to remove data as that will mess with iteration
            # and once we are done iterating this data will be erased anyways
            # lc_inbound_data.remove(data)


# this function is responsible for taking in messages and piping them into _inbound_data
def _inbound_data_listener():
    '''
    A function responsible for listening for data as soon as it gets it, and piping it into _inbound_data
    '''
    while True:
        data = _main_client.recv(1024) # reccieve data
        _inbound_data.append(data) # pipe data in

        ## OLD CODE
        # data = _main_client.recv(1024)
        # try:
        #     decoded = _decrypt_private(data)
        # except Exception as e:
        #     # print('packet loss error:', e)
        #     continue # packet loss
        # if decoded == '000':
        #     continue
        # # print('adding message to queue:', len(data_queue) + 1)
        # # print('recieved packet:', decoded)
        # data_queue.append(decoded)
        # # print('message queue:', data_queue)








####################################################################################################################################
####################################################################################################################################
####################################################################################################################################
####################################################################################################################################
####################################################################################################################################

'''
below is random code that I am not sure what to do with, so for now they remain as comments!
'''


# a function to fully recieve the message from server (to try and prevent loss)
# def full_recieve(client: socket.socket) -> str:
#     message_length = len(client.recv(1024).decode('utf-8'))
#     incoming_message = ''
#     local_length = 0
#     while local_length <= message_length:
#         incoming_message += client.recv(1024).decode('utf-8')
#         local_length = len(incoming_message)
#     return incoming_message


# def send_file(file, recieve: bool = False) -> any:
#     '''
#     A general tool function for sending files to the recipient (server, other client, etc)
#     '''
#     client = _main_client
#     encoded_file =


# def target_client(client_ip: str, client_port: int, mode: str) -> bool:
#     '''
#     Takes in the target clients ip and port, and will attempt to connect to them. If this fails, 
#     then it is possible the other client is not available.
#     This function returns a boolean, telling you whether it worked or not.
#     '''
#     global _HOST, _PORT, _valid_ports
#     global _main_client
#     global _do_print
#     # reset main_client
#     _main_client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
#     # overwrite host
#     _HOST = client_ip
#     # overwrite valid ports list
#     override_ports([client_port])
#     # overwrite port
#     _PORT = _valid_ports[0]
#     # setup other vars
#     save = _do_print

#     for i in range(30):
#         print(f'attempt {i}')
#         disable_print()
#         _main_client, _PORT = _cycle_port(_main_client)
#         _do_print = save
#         if _connected == True:
#             return True
#         time.sleep(1)
#     return False


# def _data_queue_cleaner():
#     global data_queue
#     while True:
#         if len(data_queue) > 25: # currently hard set value
#             try:
#                 latest = data_queue[-1]
#                 # latest = data_queue[-5:]
#                 # print('saving latest:', latest)
#                 data_queue = [latest]
#                 # data_queue = latest
#                 # print('set new message queue to:', data_queue)
#                 # data_queue = []
#                 # print('cleaning data queue')
#             except Exception as e:
#                 # print('clean error:', e)
#                 pass


'''
this is the parser file, which includes all of the tools for managing messages.
The parser handles adding and removing the padding in messages, and isolating full message packets
relatively quickly. It also has a variety of customization.
'''

# included modules
# from ast import literal_eval

####################################################

# this is the main class for the parser 
class Parser:
    def __init__(self): 
        # end marker for message
        self.byteEndMarker = b':~e~:'
        # heartbeat marker
        self.heartbeatMarker = ':~hb~:'
        # carry over from previous incomplete packets
        self.carry = b''


    # this function prepares messages to be sent
    # takes in an input of the encrypted data and returns
    # it in a string with the start marker
    def addPadding(self, message: bytes) -> bytes: return message + self.byteEndMarker


    # this function splits the messages by the start marker
    # and can optionally discard invalid endings that aren't complete
    def removePadding(self, message: bytes, remove_empty: bool = True) -> list:
        # add the previous data to current data
        stitched = self.carry + message
        # split message by end marker
        split = stitched.split(self.byteEndMarker)
        # if the end characters is the end marker, then that means
        # we only have complete messages so we can reset carry
        if stitched.endswith(self.byteEndMarker): self.carry = b''
        # otherwise, we analyze further
        else:
            # if the length of the list is 0, then we have nothing
            # if the length of the list is 1, the we only have
            # a singular incomplete packet
            if len(split) <= 1:
                # save the message to carry and return
                # empty list
                self.carry = stitched
                return []
            # else, if the last entry of the list is not empty,
            # that means there is an incomplete packet there
            elif split[-1]:
                # take the last packet which is incomplete and save to carry
                self.carry = split.pop(-1)
        # if requested, go ahead and remove all white spaces
        if remove_empty:
            index = 0
            for elem in split:
                if not elem: split.pop(index)
                index += 1
        # return our final formatted list of messages
        return split


'''
This is the pool file, which contains a variety of different types of objects used for variables or recieved data.
For example, on the client ON_MESSAGE, the data you recieve is a Pool.Message object.
'''

# included modules
from datetime import datetime

# the main class for containing all of the types of stuff
class Pool:

    # this is a class of tools
    class _Tools:
        def _format_time() -> str: return datetime.strftime(datetime.now(), "%d/%m/%Y, %H:%M:%S")
        

    # this is a class for message objects
    class Message:
        '''
        This class represents a Lynxy.Message object, which is what is
        used for sending data from one machine to the other. The Message object
        has 4 attributes:

        content: any
        - the actual data that was meant to be sent

        created_at: str
        - the timestamp for when the message object was created on the sending side

        recieved_at: str
        - the timestamp for when the message object was recieved on the recieving side

        public_key: rsa.PublicKey
        - the public key used to encrypt this message object when sent
        '''
        def __init__(self, data):
            self.content = data
            self.created_at = Pool._Tools._format_time()
            self.recieved_at = None # set on recieving end


from src import lynxy
import pyautogui

client = lynxy.Lynxy(bind=True)
target = ['192.168.68.113', 56774]
print('host:', client.get_host())
print('target:', target)
client.connect(target)
print('connected:', client.get_actual_target())

@client.event(lynxy.Constants.Event.ON_MESSAGE)
def on_message(data: lynxy.Pool.Message):
    print('recv:', data.content)
    # pyautogui.press(data.content)

@client.event(lynxy.Constants.Event.ON_CLOSE)
def on_close(error: lynxy.Exceptions.BaseLynxyException):
    print(error)

while True: 
    input('- >')
    client.close()


import socket
import random
import time
import threading

valid_ports = [
    11111,
    12111,
    11211,
    11121,
    11112
]


# set limit for server instances
INSTANCE_LIMIT = 5
initial_limit_print = False
alive_bound_instance = 0
alive_searching_instance = 0

# create server object, set port
#main_server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
# port = 11111
# port_iter = 0 
# valid_ports = v.valid_ports
not_inclusive_max_port_amount = len(valid_ports) - 1
port = valid_ports[0]

# set up client dict, storing name to ip correlation
client_dict = {
    'default': 0
}

unique_server_instance_dict = {}
message_lookup = {}




# def recv_all(client_socket, n): ################################################
#     # Helper function to receive n bytes or return None if EOF is hit
#     data = bytearray()
#     while len(data) < n:
#         packet = client_socket.recv(n - len(data))
#         if not packet:
#             return None
#         data.extend(packet)
#     return data





# connect to client
def connect_to_client(server: socket.socket, return_port: bool = False) -> list[socket.socket, str, int]:
    global port
    local_port = port
    # bind to first incoming ip, and port
    #server.bind(('', port))
    #print(f'server binded to port {port}')
    while True:
        try:
            print(f'[unbound] attempting to bind to port:', local_port)
            server.bind(('', local_port))
            print(f'[{local_port}] server binded to port {local_port}')
            break
        except:
            if valid_ports.index(local_port) < not_inclusive_max_port_amount:
                print('[unbound] port bind failed, cycling')
                local_port = valid_ports[valid_ports.index(local_port) + 1]
                print('[unbound] swapped to port', local_port)
            else:
                print('[unbound] all ports are being used, cancelling attempt')
                exit()
    print(f'[{local_port}] cycling port ({port}) to local_port ({local_port})')
    port = local_port

    # listen for an incoming connection
    print(f'[{local_port}] server listening for a connection...')
    server.listen(5)

    # set client, client address
    client, addr = server.accept()
    print(f'[{local_port}] connected to client at address {addr}')




    # waits for verification message
    msg = client.recv(1024).decode('utf-8')
    # msg = recv_all(client, 4) ################################################
    # message_length = int.from_bytes(msg, byteorder='big') ################################################
    # msg = recv_all(client, message_length).decode('utf-8') ################################################




    if 'verify' in msg:
        print(f'[{local_port}] recieved "verify" message, sending "verify_confirm"')
        msg = "verify_confirm".encode('utf-8')
        client.sendall(msg)
        print(f'[{local_port}] message sent')
        if return_port == True:
            return [client, addr, local_port]
        else:
            return [client, addr]

# intake username data
# def recieve_username_data(client: socket.socket, msg: str) -> str:
#     # msg = client.recv(1024).decode('utf-8')
#     split_msg = msg.split()
#     if msg:
#         if split_msg[0] == "username":
#             split_msg.remove("username")
#             username = "".join(split_msg)
#             print(f'username recieved: {username}')
#             return username










def check_alive() -> None:
    global unique_server_instance_dict
    import time
    while True:
        for por in valid_ports:
            try:
                data = unique_server_instance_dict[port]
                state = data[0]
                server_obj = data[1]
                client_obj = data[2]
                closed = is_socket_closed(client_obj)
                if closed == True:
                    print(f'[ALIVE SCANNER] the client for port {por} is not responsive, so the connection will be terminated.')
                    unique_server_instance_dict[por] = [False, server_obj, client_obj]
                # else:
                    # print(f'{por} is connected to a client and healthy')
            except:
                pass
                # print('not checking this port as they have not signed on yet')
        time.sleep(5)
        







# https://stackoverflow.com/questions/48024720/python-how-to-check-if-socket-is-still-connected
def is_socket_closed(sock: socket.socket) -> bool:
    try:
        # this will try to read bytes without blocking and also without removing them from buffer (peek only)
        #data = sock.recv(16, socket.MSG_DONTWAIT | socket.MSG_PEEK)
        data = sock.recv(16)
        if len(data) == 0:
            return True
        
    except BlockingIOError:
        return False  # socket is open and reading from it would block
    except ConnectionResetError:
        return True  # socket was closed for some other reason
    except Exception as e:
        #logger.exception("unexpected exception when checking if a socket is closed")
        return False
    return False
















# log username data into client_dict
#def log_username_data(username: str, address: str, c_dict: dict) -> dict:
def log_username_data(username: str, address: str, po: int) -> None:
    global client_dict
    client_dict[username] = address
    print(f'[{po}] logged username ({username})  and address ({address}) to client dict')
    #return client_dict














def answer_request_by_username(client: socket.socket, client_addr: str, server_po: int, c_dict: dict, cmd: str) -> None:
    # cmd = client.recv(1024).decode('utf-8')
    # split_cmd = cmd.split()
    if cmd:
        # if split_cmd[0] == 'request_ip_by_user':
        #     split_cmd.remove('request_ip_by_user')
            # username = "".join(split_cmd)
            username = cmd
            print(f'[{server_po}] client requested data associated with "{username}", searching for in client_dict...')
            try:
                req_details = str(c_dict[username])
                print(f'[{server_po}] data acquired by username: {req_details}')
                msg = req_details.encode('utf-8')
                # client.sendall(msg)
                client.sendto(msg, client_addr)
                print(f'[{server_po}] address sent back to client')
            except:
                print(f'[{server_po}] username does not exist in database, returning None.')
                msg = 'null'.encode('utf-8')
                client.sendall(msg)






# def log_message(client: socket.socket, sport: int) -> None:
#     global message_lookup
#     client.settimeout(5)
#     while True:
#         try:
#             msg = client.recv(1024).decode('utf-8')
#             message_lookup[sport] = msg
#             print('picked up and logged msg')
#             print('-------------------', msg)
#         except:
#             # print('timed out')
#             pass







# MAIN MESSAGE HANDLER
def input_handler(client: socket.socket, client_address: str, server: socket.socket, server_port: int) -> None:
    global client_dict, alive_bound_instance #, message_lookup
    print('----------------------')
    print(f'[{server_port}] INPUT HANDLER STARTED')
    print('----------------------')
    print(f'[{server_port}] starting alive check loop...')
    threading.Thread(target=lambda:check_alive(), daemon=True).start()
    #threading.Thread(target=lambda:log_message(client, server_port), daemon=True).start()
    while True: 
        if unique_server_instance_dict[server_port][0] == False:
            print(f'[{server_port}] this server on port {server_port} has no active client, and will therefore terminate its process')
            # try:
            #     server.shutdown(socket.SHUT_RDWR)
            #     client.shutdown(socket.SHUT_RDWR)
            # except:
            #     pass
            # server.detach()
            # client.detach()
            server.close()
            client.close()
            alive_bound_instance -= 1
            break
        try:
            msg = client.recv(1024).decode('utf-8')
        except:
            print(f'[{server_port}] client has been closed on port {server_port}, server instance terminating')
            # try:
            #     server.shutdown(socket.SHUT_RDWR)
            #     client.shutdown(socket.SHUT_RDWR)
            # except:
            #     pass
            # server.detach()
            # client.detach()
            server.close()
            client.close()
            alive_bound_instance -= 1
            break
        # try:
            #msg = message_lookup[server_port]
        if msg:
            split_msg = str(msg).split()
            print('---------------- MSG', msg)
            print('----------------SPLIT MSG', split_msg)
            prefix = split_msg[0]
            if prefix == 'username':
                print(f'[{server_port}] -- username prefix detected')
                split_msg.remove(prefix)
                username = "".join(split_msg)
                #client_dict = log_username_data(username, client_address, client_dict)
                log_username_data(username, client_address, server_port)
                # message_lookup[server_port] = ''
            elif prefix == 'request_ip_by_user':
                print(f'[{server_port}] -- request_ip_by_user prefix detected')
                split_msg.remove(prefix)
                username = "".join(split_msg)
                answer_request_by_username(client, client_address, server_port, client_dict, username)
                # message_lookup[server_port] = ''
        # except:
            # pass
    exit()







def start_server_instance(server: socket.socket, instance_at_start: int) -> None:
    global alive_searching_instance, alive_bound_instance, unique_server_instance_dict
    print(f'[SERVER STARTER] starting a new searching instance with num: {instance_at_start}')
    main_client, client_address, server_port = connect_to_client(server, return_port=True)
    print('[SERVER STARTER] binded to client, decreasing search and increasing bound')
    alive_searching_instance -= 1
    alive_bound_instance += 1
    print('[SERVER STARTER] bound is now:', alive_bound_instance)
    print('[SERVER STARTER] signing in to active dict with port, logging True and main_client')
    print(f'[SERVER STARTER] client address, server port: {client_address}, {server_port}')
    # active_instance_dict[server_port] = [True, main_client]
    unique_server_instance_dict[server_port] = [True, server, main_client]
    # print('logging client address to detect if closed later')
    # unique_server_instance_dict[main_client] = True
    print('[SERVER STARTER] transferring to input handler') 
    input_handler(main_client, client_address, server, server_port)







def start_instance_handler() -> None:
    global alive_searching_instance, alive_bound_instance, initial_limit_print
    while True:
        if alive_searching_instance == 0:
            if alive_bound_instance < INSTANCE_LIMIT:
                initial_limit_print = False
                print(f'[SERVER HANDLER] starting instance, searching count is {alive_searching_instance} (before)')
                alive_searching_instance += 1
                print(f'[SERVER HANDLER] starting instance, searching count is {alive_searching_instance} (after)')

                main_server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                # main_server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1) # 1
                threading.Thread(target=lambda:start_server_instance(main_server, alive_bound_instance + 1), daemon=True).start()
            else:
                if initial_limit_print == False:
                    time.sleep(3) # time for other things to calm down
                    print('[SERVER HANDLER] INSTANCE LIMIT REACHED, NO MORE INSTANCES WILL BE MADE')
                    print('[SERVER HANDLER] BOUND COUNT:', alive_bound_instance)
                    initial_limit_print = True









# start normal handler
start_instance_handler()


import socket
import random
import time
import threading

# all valid ports to cycle through
valid_ports = [
    11111,
    12111,
    11211,
    11121,
    11112
]

# set limit for server instances
INSTANCE_LIMIT = 5
initial_limit_print = False
alive_bound_instance = 0
alive_searching_instance = 0

# set port limits, defaults
not_inclusive_max_port_amount = len(valid_ports) - 1
default_port = valid_ports[0]

# set up client dict, storing username to ip correlation
client_dict = {
    'default': 0
}

# a dictionary where a port number is associated with [State of living (bool), server_socket, client_socket] NOTE: server_socket is not used
unique_server_instance_dict = {}
for port in valid_ports:
    unique_server_instance_dict[port] = [False, 'x', 'x']





### FUNCTIONS

## TOOlS

# a function that is ran and evaluates all sockets, killing them if neccesary (thread) (ONE OF THE FEW FUNCTIONS WITH GLOBAL ACCESS TO VARS)
def check_alive() -> None:
    global unique_server_instance_dict
    import time
    while True:
        for port in valid_ports:
            try:
                data = unique_server_instance_dict[port]
                state = data[0]
                server_obj = data[1]
                client_obj = data[2]
                closed = is_socket_closed(client_obj)
                if closed == True:
                    print(f'[ALIVE SCANNER] the client for port {port} is not responsive, so the connection will be terminated.')
                    unique_server_instance_dict[port] = [False, server_obj, client_obj]
                    print(f'[ALIVE SCANNER] closing associated server and client...')
                    server_obj.close()
                    client_obj.close()
                # else:
                    # print(f'{por} is connected to a client and healthy')
            except:
                pass
                # print('not checking this port as they have not signed on yet')
        time.sleep(5)



# A function to check if the socket is closed (not sure how it works, though)
# https://stackoverflow.com/questions/48024720/python-how-to-check-if-socket-is-still-connected
def is_socket_closed(sock: socket.socket) -> bool:
    try:
        # this will try to read bytes without blocking and also without removing them from buffer (peek only)
        data = sock.recv(16, socket.MSG_DONTWAIT | socket.MSG_PEEK)
        # data = sock.recv(16)
        if len(data) == 0:
            return True
        
    except BlockingIOError:
        return False  # socket is open and reading from it would block
    except ConnectionResetError:
        return True  # socket was closed for some other reason
    except Exception as e:
        #logger.exception("unexpected exception when checking if a socket is closed")
        return False
    return False









## CORE FUNCTIONS
# connect to client
def connect_to_client(server: socket.socket, port_override: int = 0, return_port: bool = False) -> list:
    # access global
    #local_port = default_port


    try:
        # iterate through valid ports, check if any are not alive in unique_server_instance_dict
        found_open = False
        if port_override == 0:
            for port in valid_ports:
                data = unique_server_instance_dict[port]
                is_alive = data[0]
                server_obj = data[1]
                client_obj = data[2]
                if is_alive == False:
                    found_open = True
                    open_port = port
                    break
            if found_open == True:
                print(f'[unbound] open port found: {open_port}')
                #server.bind(('', open_port))
                server.bind(('localhost', open_port))
                print(f'[{open_port}] server binded to port {open_port}')
            else:
                print('[unbound] all ports are being used, cancelling attempt')
                exit()
        else:
            open_port = port_override
            print(f'[unbound] port overrided to: {open_port}')
            #server.bind(('', open_port))
            server.bind(('localhost', open_port))
            print(f'[{open_port}] server binded to port {open_port}')
    except Exception as e:
        print(f'[unbound] an error occured when trying to bind to ports: {e}')



    # print(f'[{local_port}] cycling port ({default_port}) to local_port ({local_port})')
    # default_port = local_port

    # listen for an incoming connection
    print(f'[{open_port}] server listening for a connection...')
    server.listen(5)

    # set client, client address
    client, client_address = server.accept()
    print(f'[{open_port}] connected to client at address {client_address}')

    # waits for verification message
    msg = client.recv(1024).decode('utf-8')

    # if verification message contains "verify", return "verify_confirm"
    if 'verify' in msg:
        print(f'[{open_port}] recieved "verify" message, sending "verify_confirm"')
        msg = "verify_confirm".encode('utf-8')
        client.sendall(msg)
        print(f'[{open_port}] verify_confirm sent')
        if return_port == True:
            return [client, client_address , open_port] # if port requested, return client, client_address, and the port used for this instance
        else:
            return [client, client_address] # if no port requested, return client, client_address
        


# logs username data into client_dict (ONE OF THE FEW FUNCTIONS WITH GLOBAL ACCESS TO VARS)
def log_username_data(username: str, address: str, port: int) -> None:
    global client_dict
    client_dict[username] = address
    print(f'[{port}] logged username ({username})  and address ({address}) to client dict')



# answer requests with a username with the data assigned to that username -> ip, port
def answer_request_by_username(client: socket.socket, client_address: str, port: int, client_dict: dict, username: str) -> None:
    # if string not empty, continue
    if username:
        print(f'[{port}] client requested data associated with "{username}", searching for in client_dict...')
        # try to search for that username in the databse
        try:
            req_details = str(client_dict[username])
            print(f'[{port}] data acquired by username: {req_details}')
            msg = req_details.encode('utf-8')
            # client.sendall(msg)
            #client.sendto(msg, client_address)
            print(f'[{port}] address found')

        # if failed, then there is no person of that username, send that back
        except:
            print(f'[{port}] username does not exist in database, returning null')
            msg = 'null'.encode('utf-8')
        client.sendall(msg)
        print(f'[{port}] request results sent back to client')



# handles inputs while scanning for messages (ONE OF THE FEW FUNCTIONS WITH GLOBAL ACCESS TO VARS)
def input_handler(client: socket.socket, client_address: str, server: socket.socket, server_port: int) -> None:
    global client_dict, alive_bound_instance
    # starts
    print(f'----------------------\n[{server_port}] INPUT HANDLER STARTED\n----------------------')

    # starts thread to check for what sockets are alive (disabled, currently) 
    print(f'[{server_port}] starting alive check loop...') ################################################
    threading.Thread(target=lambda:check_alive(), daemon=True).start() ################################################

    # internal loop
    while True: 
        # see if the current server is supposed to be alive
        is_alive = unique_server_instance_dict[server_port][0]
        if is_alive == False:
            print(f'[{server_port}] this server on port {server_port} has no active client, and will therefore terminate its process')
            # close client and server if this port is not connected to a client
            server.close()
            client.close()
            # lower how many are considered alive and bound so a new instance can be made 
            alive_bound_instance -= 1
            break

        # try to recieve a message from the client
        try:
            msg = client.recv(1024).decode('utf-8')

        # failed due to lack of connection, or whatnot -> close server and client
        except Exception as e:
            # print(f'[{server_port}] the client on port {server_port} has been closed, and the server will therefore terminate its process')
            print(f'[{server_port} - ERROR] server has recieved an error, terminating: {e}')
            server.close()
            client.close()
            # lower how many are considered alive and bound so a new instance can be made 
            alive_bound_instance -= 1
            break
        
        # if message is not empty, continue
        if msg:
            # split message into a list, see command (list[0], assigned to "prefix")
            split_msg = str(msg).split()
            print('---------------- MSG:', msg)
            print('----------------SPLIT MSG:', split_msg)
            prefix = split_msg[0]

            # if prefix is username, remove the prefix from the string and re-connect it together
            if prefix == 'username':
                print(f'[{server_port}] -- username prefix detected')
                split_msg.remove(prefix)
                username = "".join(split_msg)
                # call on the function to log username data, passing in -> username, client_address, server_port (just for printing)
                log_username_data(username, client_address, server_port)

            # elif prefix is request_ip_by_user, remove the prefix from the string and re-connect it together
            elif prefix == 'request_ip_by_user':
                print(f'[{server_port}] -- request_ip_by_user prefix detected')
                split_msg.remove(prefix)
                username = "".join(split_msg)
                # call on the function to answer this request, passing in -> client, client_address, server_port, client_dict, username
                answer_request_by_username(client, client_address, server_port, client_dict, username)

    # exit if the code broke out of the loop
    exit()



# this function is called on to set up the initial server instance (thread) (ONE OF THE FEW FUNCTIONS WITH GLOBAL ACCESS TO VARS)
def start_server_instance(instance_at_start: int) -> None:
    global alive_searching_instance, alive_bound_instance, unique_server_instance_dict

    # initially make server binded before loop
    server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # server = socket.socket(socket.AF_INET, socket.TCP_NODELAY)

    # set the client object, the clients address, and the servers binded port from connect_to_client, passing in a server socket
    print(f'[SERVER STARTER-{instance_at_start}] starting a new searching server instance with num: {instance_at_start}')
    client, client_address, server_port = connect_to_client(server, return_port=True)
    print(f'[SERVER STARTER-{instance_at_start}] server instance {instance_at_start} binded to client')

    # sign into the dictionary, passing in -> alive or dead (bool, default is True as we are starting), server socket, client_socket
    print(f'[SERVER STARTER-{instance_at_start}] client address, server port: {client_address}, {server_port}')
    unique_server_instance_dict[server_port] = [True, server, client]
    print(f'[SERVER STARTER-{instance_at_start}] signing in to active dict with port, logging True and main_client')

    # lower searching as server instance has found client, increase bound (will trigger a new instance creation)
    print(f'[SERVER STARTER-{instance_at_start}] decreasing search and increasing bound')
    alive_searching_instance -= 1
    alive_bound_instance += 1
    print(f'[SERVER STARTER-{instance_at_start}] searching is now:', alive_searching_instance)
    print(f'[SERVER STARTER-{instance_at_start}] bound is now:', alive_bound_instance)

    # start the input handler, this thread has now finished its task
    print(f'[SERVER STARTER-{instance_at_start}] transferring to input handler') 
    input_handler(client, client_address, server, server_port)



# the manager for starting server instances (ONE OF THE FEW FUNCTIONS WITH GLOBAL ACCESS TO VARS)
def start_instance_handler() -> None:
    global alive_searching_instance, alive_bound_instance, initial_limit_print
    
    # main loop
    while True:
        # if no instances are searching, continue
        if alive_searching_instance == 0:
            # if the amount of bound instances is below the limit, start a new instance (assigned a cosmetic num which is alive_bound_instance + 1)
            if alive_bound_instance < INSTANCE_LIMIT:
                initial_limit_print = False
                # increment alive_bound_instance
                print(f'[SERVER HANDLER] starting instance, searching count is {alive_searching_instance} (before)')
                alive_searching_instance += 1
                print(f'[SERVER HANDLER] starting instance, searching count is {alive_searching_instance} (after)')
                
                # create a var, then immediately start, this gets overwritten when a new one is created
                new_server_thread = threading.Thread(target=lambda:start_server_instance(alive_bound_instance + 1), daemon=True)
                new_server_thread.start()

            # if we are the limit, print taht we are (initial_limit_print used to not spam console) and finish
            else:
                if initial_limit_print == False:
                    time.sleep(3) # time for other things to calm down
                    print(f'[SERVER HANDLER] INSTANCE LIMIT REACHED, NO MORE INSTANCES WILL BE MADE\n[SERVER HANDLER] BOUND COUNT: {alive_bound_instance}')
                    initial_limit_print = True




# the only acting line, calling on the handler which handles everything
start_instance_handler()


'''
This is the exceptions file, which contains all of the Exceptions that are raised throughout the code.
These are helpful to know in case you want to catch any errors when doing a variety of different functions.
'''

# this is the class for all errors
# which contains all other classes of errors
class Exceptions:

    class BaseLynxyException(Exception): pass


    class ClientNotConnectedError(BaseLynxyException):
        def __init__(self, message='The client is not connected.'):
            self.message = message
            super().__init__(message)

    class NoExteralPublicKeyError(BaseLynxyException):
        def __init__(self, message='External public key has not been loaded.'):
            self.message = message
            super().__init__(message)

    
    class EmptyDataError(BaseLynxyException):
        def __init__(self, message='The data you attempted to send is empty.'):
            self.message = message
            super().__init__(message)


    class SendingTimeoutError(BaseLynxyException):
        def __init__(self, message='The client timed out waiting to send data (send lock).'):
            self.message = message
            super().__init__(message)


    class ConnectionFailedError(BaseLynxyException):
        def __init__(self, message=None):
            self.message = message
            super().__init__(message)


# this below code has been contributed to by chat gpt
import socket
import time

valid_ports = [
    11111,
    12111,
    11211,
    11121,
    11112,
    22111,
    12211,
    11221,
    11122
]

# define all global vars
HOST, PORT = '', valid_ports[0] # localhost
main_client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)



## FUNCTIONS
# cycles port connection
def cycle_port(client: socket.socket) -> socket.socket:
    connected = False
    for port in valid_ports:
        try:
            print(f'[PORT CYCLE] Client trying port: {port}')
            client.connect((HOST, port))
            print(f'[PORT CYCLE] Client connected to: {port}')
            print('----------------------------------------------')
            connected = True
            break
        except IndexError:
            port = valid_ports[0]
            print(f'[PORT CYCLE - RESET 1] Client resetting port to: {port}')
        except:
            try:
                print(f'[PORT CYCLE] Client port cycling: {port} -> {valid_ports[valid_ports.index(port) + 1]}')
            except IndexError:
                port = valid_ports[0]
                print(f'[PORT CYCLE - RESET 2] Client resetting port to: {port}')
    if connected == True:
        return client, port
    else:
        print('[PORT CYCLE] the client can not find a open valid server port, exiting')
        exit()



# a function to fully recieve the message from server (to try and prevent loss)
def full_recieve(client: socket.socket) -> str:
    message_length = len(client.recv(1024).decode('utf-8'))
    incoming_message = ''
    local_length = 0
    while local_length <= message_length:
        incoming_message += client.recv(1024).decode('utf-8')
        local_length = len(incoming_message)
    return incoming_message



# a function for submitting username data to the server
def submit_username_data(client: socket.socket, message: str) -> None:
    encoded_message  = message.encode('utf-8')
    client.sendall(encoded_message)
    print(f"Sent:     {message}")
    incoming_data = client.recv(1024).decode('utf-8')
    print(f"Received: {incoming_data}")


# requests ip and port from server
def request_data(client: socket.socket, message: str) -> socket.socket:
    encoded_message = message.encode('utf-8')
    client.sendall(encoded_message)
    print(f"Sent:     {message}")
    # incoming_data = full_recieve(client)
    incoming_data = client.recv(1024).decode('utf-8')
    print(f"Received: {incoming_data}")



# a general message sender
def general_send(client: socket.socket, message: str) -> None:
    encoded_message = message.encode('utf-8')
    client.sendall(encoded_message)
    print(f"Sent:     {message}")
    # incoming_data = full_recieve(client)
    incoming_data = client.recv(1024).decode('utf-8')
    print(f"Received: {incoming_data}")




# establish the connection to a port that the server is on
main_client, PORT = cycle_port(main_client)

# next, send a send a message to the server
submit_username_data(main_client, 'username SketchedDoughnut')

# necessary delay
time.sleep(1)

# next, request username data
request_data(main_client, 'request_by_user SketchedDoughnut')

# necessary delay
time.sleep(1)

# finally, end the session
general_send(main_client, 'end_session')


import socket
import time
import threading
import json

valid_ports = [
    11111,
    12111,
    11211,
    11121,
    11112
]

# set up server ip, port to connect to
server_ip = input('enter ip: ')
# valid_ports = v.valid_ports
not_inclusive_max_port_amount = len(valid_ports) - 1
server_port = valid_ports[0]

# start client timeout
# connection_timeout = False








# def timeout_connection(timeout_time: int = 5) -> None:
#     global connection_timeout
#     init_time = time.time()
#     while True:
#         new_time = time.time()
#         dif = new_time - init_time
#         if dif >= timeout_time:
#             connection_timeout = True
#             print('timed out')
#             exit()









# def info_connect() -> dict:
#     # attempt to connect to the data server object
#     info_client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
#     info_client.connect((server_ip, 22222))
#     active_data = info_client.recv(1024).decode('utf-8')

#     active_data = active_data.strip('{}')
#     # Split by commas to get individual key-value pairs
#     pairs = active_data.split(',')

#     # Initialize an empty dictionary to store the parsed data
#     data_dict = {}

#     # Iterate over each pair and split by colon to get key and value
#     for pair in pairs:
#         key, value = pair.split(':')
#         # Strip spaces and convert key to integer
#         key = int(key.strip())
#         # Evaluate value and store it in the dictionary
#         value = eval(value.strip())
#         data_dict[key] = value

#     return data_dict




# connects to server
def connect_to_server(client: socket.socket):
    global server_port, connection_timeout
    '''sets up the initial connection to the server, verifies it can send data over'''
    
    # connect to server ip, server port
    print('cycling connection to server until succeed / fail...')
    attempt_count = 0
    while True:
        print(f'attempt: {attempt_count}')
        try:
            print('trying port:', server_port)
            # con_thread = threading.Thread(target=lambda:client.connect((server_ip, server_port)))
            # con_thread.start()
            # tm_thread = threading.Thread(target=lambda:timeout_connection())
            # tm_thread.start()
            # if connection_timeout == True:
            #     print('timed out - main')
            # if con_thread.is_alive() == False:
            #client.connect((server_ip, server_port))
            client.connect(('localhost', server_port))
            break
        except Exception as e:
            # print('error:', e)
            # exit()
            attempt_count += 1
            if attempt_count > 0:
                print('failed, moving to next port')
                if valid_ports.index(server_port) < not_inclusive_max_port_amount:
                    server_port = valid_ports[valid_ports.index(server_port) + 1]
                    print('swapped to port', server_port)
                else:
                    print('all ports exhausted, exiting')
                    exit()
                attempt_count = 0
            # attempt_count += 1
            # if attempt_count > 9:
            #     print('connection failed 10 times, exiting')
            #     exit()

    # loop until get verify back
    while True:
        msg = "verify"
        # message_length = len(msg) ################################################ 
        # client.sendall(message_length.to_bytes(4, byteorder='big'))  # Send the length of the message ################################################
        client.sendall(msg.encode('utf-8')) # send message  
        print('sent verify message, waiting for return')
        incoming = client.recv(1024).decode('utf-8') # recieve message 
        if incoming == 'verify_confirm': # message to confirm verify = "verify_confirm"
            print('recieved verify_confirm')
            break # breaks out of loop when it recieves
    print('connection made, verified connection')









def submit_username_data(client: socket.socket, username: str) -> None:
    '''submits username to the server that gets associated with this clients ip address
    NOTE: USERNAME CAN HAVE NO SPACES, ANY SPACES WILL BE REMOVED'''
    msg = f'username {username}'.encode('utf-8')
    # message_length = len(msg) # chatty
    # client.sendall(message_length.to_bytes(4, byteorder='big')) # chatty
    client.sendall(msg)
    print(f'submitted username to server: {username}')










def request_by_username(client: socket.socket, username: str) -> tuple:
    '''requests an address from the server, getting back an ip address and port'''
    msg = f'request_ip_by_user {username}'.encode('utf-8')
    # print('senttoff msg:', msg)
    client.sendall(msg)
    target_data = client.recv(1024).decode('utf-8')
    if target_data == 'null':
        print('request failed, there is not currently anyone using that username')
        return ('x', 0)
    else:
        # ????????????????????????????????????????
        target_data = target_data.strip('()')
        target_ip, target_port = target_data.split(',')
        target_ip = target_ip.strip().strip("'")
        target_port = target_port.strip()
        target_port = int(target_port)
        print(f'acquired ip and port of {username}, respectively: {target_ip}, {target_port}')
        return (target_ip, target_port)







# port = valid_ports[0]
for port in valid_ports:
    server_port = port
    main_client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    print('CLIENT: CONNECTING TO PORT', port)
    connect_to_server(main_client)
    submit_username_data(main_client, 'SketchedDoughnut')
    time.sleep(1) # NECESSARY OR MESSAGES GET SCRAMBLED AND CUT OFF BY EACHOTHER
    target_ip, target_port = request_by_username(main_client, 'SketchedDoughnut')
    time.sleep(1)


'''
This is the constants file, which contains all of the constants, either as Enums or literals.
They are used often in function inputs to signifiy different things that are done.
'''

# included modules
from enum import Enum

# the main class for containing all of the Enums
# basically fancy constants that keep their color
class Constants:
    
    # class for all events
    class Event(Enum):
        '''
        This class contains all constants pertaining to events.
        '''
        ON_MESSAGE = 'on message'
        ON_CLOSE = 'on close'
        ON_CONNECT = 'on connect'

    
    # class for all connection types
    class ConnectionType(Enum):
        '''
        This class contains all constants pertaining to different connection types
        '''
        ERROR = 'error'
        EVENT = 'event'
        NONE = 'none'


    # class for all connection biases
    class ConnectionBias:
        '''
        This class contains all constants pertaining to the bias of who goes first when connecting with Lynxy
        '''
        FIRST = True
        LAST = False
        NONE = None


from src import lynxy

# set up lynxy client
client = lynxy.Lynxy(bind=True)
client.set_connection(lynxy.Constants.ConnectionType.ERROR)
print(client.get_host())

source = r'D:\VScode\github practice\filter\original\die.mp4'

target = ['192.168.68.113', 56774]
client.connect(target)

f = open(source, 'rb')
contents = f.read()
f.close()

client.send(contents)

client.close()


# from src import lynxy
import lynxy
from rich import print

inst = lynxy.Lynxy(bind=True)
print('initialized')
host = inst.get_host()
target = ('192.168.68.129', 56774)

print(f'host: {host}')
print(f'target: {target}')

inst.connect(target)
print('connected')

@inst.event(lynxy.Constants.Event.ON_MESSAGE)
def recv(msg: lynxy.Pool.Message): 
    print(msg.content)

@inst.event(lynxy.Constants.Event.ON_CLOSE)
def close(msg):
    print('connection closed:', msg)

while True:
    msg = input('-> ')
    if msg == 'exit': 
        inst.close()
        break
    inst.send(msg)


'''
This is the sec file, which contains all of the encryption tools. There are functions for both encryption and decryption
for RSA and Fernet (AES with a 128 bit key). RSA is only used for exchanging Fernet keys when the two machines
connect to each other.
'''

# included modules
import pickle

# external modules
import rsa
from cryptography.fernet import Fernet

# files
from .exceptions import Exceptions

####################################################

# the main class for managing
# security layers
class Sec:
    def __init__(self):
        ############## 
        # FOR RSA ENCRYPTION
        # AKA ASYMMETRICAL
        ##############
        # represents internal public and private keys
        # for this end of communication
        self.int_pub_key, self.int_priv_key = self._gen_access_keys()
        # represents the external public key, we don't get their private key
        # for other end of communication
        self.ext_pub_key = None
        ##############
        # FOR FERNET ENCRYPTION
        # AKA SYMMETRICAL
        ##############
        # represents the main fernet object and key
        # this manager by default starts off with its own encryption key and tool
        self.fernet_key = Fernet.generate_key()
        self.fernet_tool = Fernet(self.fernet_key)

    
    # generates public/private keys for encryption
    def _gen_access_keys(self) -> tuple[rsa.PublicKey, rsa.PrivateKey]: return rsa.newkeys(1024)


    # function that loads RSA public key
    def load_RSA(self, ext_pub_key: rsa.PublicKey) -> None: self.ext_pub_key = ext_pub_key


    # function that loads Fernet key
    def load_Fernet(self, fernet_key: bytes) -> None: 
        self.fernet_key = fernet_key
        self.fernet_tool = Fernet(self.fernet_key)
        return None


    # function that encrypts using RSA and a public key
    def RSA_encrypt(self, data: any, use_internal: bool = False) -> bytes:
        # if use_internal, then we use int_pub_key to encrypt
        # otherwise we use ext_pub_key
        if use_internal: activeKey = self.int_pub_key
        else: activeKey = self.ext_pub_key
        if activeKey == None: raise Exceptions.NoExteralPublicKeyError()
        pickledData = pickle.dumps(data)
        encryptedData = rsa.encrypt(pickledData, activeKey)
        return encryptedData


    # function that decrypts using RSA and a private key
    def RSA_decrypt(self, data: bytes) -> any:
        decryptedData = rsa.decrypt(data, self.int_priv_key)
        unpickledData = pickle.loads(decryptedData)
        return unpickledData
    

    # function that encrypts using Fernet and a key
    def Fernet_encrypt(self, data: any) -> bytes:
        pickledData = pickle.dumps(data)
        encryptedData = self.fernet_tool.encrypt(pickledData)
        return encryptedData
    

    # function that decrypts using Fernet and a key
    def Fernet_decrypt(self, data: bytes) -> any:
        decryptedData = self.fernet_tool.decrypt(data)
        unpickledData = pickle.loads(decryptedData)
        return unpickledData


'''
This is the Lynxy client package, which allows communication with other Lynxy machines.
It simplifies a lot of the things such as security, ease of communication, and more. 
To find documentation, go to the github~!
- Github: https://github.com/SketchedDoughnut/lynxy
'''

# extending main lynx file
from .lynx import *


import socketserver

# https://nora.codes/tutorial/socketserver-the-python-networking-module-you-didnt-know-you-needed/
class CaaSHandler(socketserver.StreamRequestHandler):
    def handle(self):
        self.wfile.write(b"Enter some data to be capitalized:\n")
        data = self.rfile.readline()
        self.wfile.write(data.upper())

if __name__ == "__main__":
    server = socketserver.TCPServer(('', 11111), CaaSHandler)
    server.serve_forever()


'''
https://pymotw.com/2/SocketServer/
https://stackoverflow.com/questions/37689518/python-socketserver-or-socket-with-thread
https://docs.python.org/3/library/socketserver.html
https://gist.github.com/pklaus/c4c37152e261a9e9331f
https://www.bogotobogo.com/python/python_network_programming_socketserver_framework_for_network_servers.php#google_vignette
https://stackoverflow.com/questions/18563664/socketserver-python
https://realpython.com/courses/python-sockets-part-1/
'''


from src import lynxy

# set up instance
inst = lynxy.Lynxy()
# inst.connect('', 0)
# inst.send('test')

# inst.close
# inst.connect
# inst.event
# inst.get_actual_target
# inst.get_host
# inst.recv
# inst.send

# set up a decorator for the on_message event
@inst.event(lynxy.Constants.Event.ON_MESSAGE)
def on_message(data: lynxy.Pool.Message):
    print(data.content)
    print(data.created_at)
    print(data.recieved_at)

# set up a decorator for the on_close event
@inst.event(lynxy.Constants.Event.ON_CLOSE)
def close(data):
    print(data)

# set up a decorator for the on_connect event
@inst.event(lynxy.Constants.Event.ON_CONNECT)
def connect(data):
    print(data)


# raw string forms
OPERATION_SUCCESS = "000"

OPERATION_FAIL = "001"

END_SESSION = "002"

INVALID_COMMAND = "003"

CRASH = "004"

INVALID_USERNAME_DATA = "100"

INVALID_AUTH_TOKEN = "101"

USER_NOT_AUTHORIZED = "102"

INVALID_MESSAGE = "103"

ALREADY_REGISTERED_USERNAME = "104"

USERNAME_EXISTS = "105"

FEATURE_DISABLED = "106"

KILL_ALL = "200"


# encoded forms
# OPERATION_SUCCESS = OPERATION_SUCCESS.encode()
# OPERATION_FAIL = OPERATION_FAIL.encode()
# END_SESSION = END_SESSION.encode()
# INVALID_COMMAND = INVALID_COMMAND.encode()
# CRASH = CRASH.encode()
# INVALID_USERNAME_DATA = INVALID_USERNAME_DATA.encode()
# INVALID_AUTH_TOKEN = INVALID_AUTH_TOKEN.encode()
# USER_NOT_AUTHORIZED = USER_NOT_AUTHORIZED.encode()
# INVALID_MESSAGE = INVALID_MESSAGE.encode()
# ALREADY_REGISTERED_USERNAME = ALREADY_REGISTERED_USERNAME.encode()
# USERNAME_EXISTS = USERNAME_EXISTS.encode()
# KILL_ALL = KILL_ALL.encode()


from src import lynxy
import keyboard

client = lynxy.Lynxy(bind=True)
target = ['192.168.68.121', 56774]
print('host:', client.get_host())
print('target:', target)
client.connect(target)
print('connected:', client.get_actual_target())

@client.event(lynxy.Constants.Event.ON_CLOSE)
def on_close(error: lynxy.Exceptions.BaseLynxyException):
    print(error)
    
while True:
    pressed = keyboard.read_key()
    print('pressed:', pressed)
    client.send(pressed)


# import socket

# ip = '' # localhost
# port = 1111

# client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
# input('input to start: ')
# client.connect((ip, port))
# client.sendall('message from client!'.encode('utf-8'))
# print(client.recv(1024).decode('utf-8'))

import socket

HOST, PORT = "localhost", 11111

for i in range(2):
    # data = f"Hello, World: {i}"
    data = f'request_by_user SketchedDoughnut-{i}'
    # Create a socket (SOCK_STREAM means a TCP socket)
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        # Connect to server and send data
        sock.connect((HOST, PORT))
        sock.sendall(data.encode('utf-8'))

        # Receive data from the server and shut down
        received = sock.recv(1024)

    print(f"Sent:     {data}")
    print(f"Received: {received.decode('utf-8')}")


import socket
# import time
import pickle
import threading

# installs
import rsa

_valid_ports = [
    11111,
    12111,
    11211,
    11121,
    11112,
    22111,
    12211,
    11221,
    11122,
    22222
]

# define all global vars
_HOST, _PORT = '', _valid_ports[0] # local_HOST
_main_client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

# override info
_ov_ports = []
_do_print = False

# status vars
_connected = False

# safety vars
_server_public_key = 0
_client_private_key = 0
_client_public_key = 0

# listener features
message_queue = []

## FUNCTIONS - overrides, features
def override_ports(ports: list) -> None:
    ''' 
    Overrides what ports the client will attempt to connect to
    '''
    global _ov_ports
    _ov_ports = ports

# disable prints
def disable_print() -> None:
    '''
    Disables the client from printing messages
    '''
    global _do_print
    _do_print = False

# enable prints
def enable_print() -> None:
    '''
    Enables the client to print messages
    '''
    global _do_print
    _do_print = True

# function to handle printing
def pprint(msg: str) -> None:
    '''
    A function meant for filtering prints based on if it is enabled or disabled - This is meant for internal use
    '''
    if _do_print:
        print(msg)
    else:
        pass

# function to display current data
def get_data() -> dict:
    '''
    Returns data about the current client in the form of a dictionary
    '''
    return {
        'client info': {
            'ip': _HOST,
            'port': _PORT
        },
        'data': {
            'message_queue': message_queue
        },
        'sillies': 'sillies :3'
    }


























## FUNCTIONS - safety
def _gen_access_keys() -> tuple:
    '''
    Generates a public and private key, and returns them in a tuple
    '''
    public_key, private_key = rsa.newkeys(1024)
    return public_key, private_key

def _handshake(client: socket.socket) -> None:
    global _client_private_key, _client_public_key, _server_public_key

    pprint('[CLIENT HANDSHAKE] Handshaking with server...')
    # generate public and private keys (OBJECTS)
    _client_public_key, _client_private_key = _gen_access_keys()
    # pprint('[HANDSHAKE]: generating public and private key')

    # get servers public key (BYTES -> OBJECT)
    pickled_server_public_key = client.recv(1024)
    _server_public_key = pickle.loads(pickled_server_public_key)
    # pprint('[HANDSHAKE]: getting server public key')

    # pickle and send our public key (OBJECT -> BYTES)
    pickled_client_public_key = pickle.dumps(_client_public_key)
    client.sendall(pickled_client_public_key)
    # pprint('[HANDSHAKE]: sending public key')

    # print('client priv key:', _client_private_key)
    # print('client pub key:', _client_public_key)
    # print('client server pub key:', _server_public_key)

def _encrypt_public(data) -> bytes:
    pickled_data = pickle.dumps(data)
    encoded_data = rsa.encrypt(pickled_data, _server_public_key)
    return encoded_data

def _decrypt_private(data) -> any:
    decrypted_data = rsa.decrypt(data, _client_private_key)
    loaded_data = pickle.loads(decrypted_data)
    return loaded_data



















## FUNCTIONS - operations
# cycles port connection
def _cycle_port(client: socket.socket) -> tuple[socket.socket, int, bool]:
    global _connected
    '''
    An internal function used to cycle through the ports in _valid_ports to try and find a connection
    '''
    _connected = False
    out_port = 0
    for port in _valid_ports:
        out_port = port
        try:
            pprint(f'[PORT CYCLE] Client trying port: {port}')
            client.connect((_HOST, port))
            pprint(f'[PORT CYCLE] Client connected to: {port}')
            pprint('----------------------------------------------')
            _connected = True
            break
        except IndexError:
            port = _valid_ports[0]
            pprint(f'[PORT CYCLE - RESET 1] Client resetting port to: {port}')
        except:
            try:
                pprint(f'[PORT CYCLE] Client port cycling: {port} -> {_valid_ports[_valid_ports.index(port) + 1]}')
            except IndexError:
                port = _valid_ports[0]
                pprint(f'[PORT CYCLE - RESET 2] Client resetting port to: {port}')
    if _connected == True:
        _handshake(client)
        return client, out_port, True
    elif _connected == False:
        pprint('[PORT CYCLE] the client can not find a open valid server port, exiting')
        return client, _PORT, False




































# a function to fully recieve the message from server (to try and prevent loss)
# def full_recieve(client: socket.socket) -> str:
#     message_length = len(client.recv(1024).decode('utf-8'))
#     incoming_message = ''
#     local_length = 0
#     while local_length <= message_length:
#         incoming_message += client.recv(1024).decode('utf-8')
#         local_length = len(incoming_message)
#     return incoming_message

# a function for submitting username data to the server
def submit_username_data(username: str) -> str:
    '''
    Submits a username to the server, which the server will associate with your IP and port.
    Returns a message that confirms that the action has happened.
    '''
    # local override for package form
    client = _main_client
    message = username
    # encoded_message = message.encode('utf-8')
    message2 = f'username {message}'
    # encoded_message = message2.encode('utf-8') # added username prefix by default
    encoded_message = _encrypt_public(message2)
    client.sendall(encoded_message)
    pprint(f"Sent:     {message2}")
    # incoming_data = client.recv(1024).decode()\
    incoming_data = client.recv(1024)
    incoming_data = _decrypt_private(incoming_data)
    pprint(f"Received: {incoming_data}")
    return incoming_data

# requests ip and port from server
def request_username_data(username: str) -> any:
    '''
    requests data associated with a username from the server, and either returns a status code, meaning you entered an invalid username, 
    or returns the IP and port of the user in a list.
    '''
    # local override for package form
    client = _main_client
    message = username
    # encoded_message = message.encode('utf-8')
    message2 = f'request_by_user {message}'
    # encoded_message = message2.encode('utf-8') # added request_by_user prefix by default
    encoded_message = _encrypt_public(message2)
    client.sendall(encoded_message)
    pprint(f"Sent:     {message2}")
    # incoming_data = full_recieve(client)
    # incoming_data = client.recv(1024).decode('utf-8')
    incoming_data = client.recv(1024)
    incoming_data = _decrypt_private(incoming_data)
    pprint(f"Received: {incoming_data}")
    if incoming_data == '100':
        return incoming_data

    # parse into list
    # address_str = incoming_data.strip('()')
    # ip_str, port_str = address_str.split(',')
    # ip_str = ip_str.strip().strip("'")
    # port_int = int(port_str.strip())
    # incoming_data = [ip_str, port_int]
    return incoming_data

# a general message sender
def send_msg(data: any, recieve: bool = True) -> any:
    '''
    A general tool function for sending messages to the recipient (server, other client, etc)
    '''
    message = data
    # local override for package form
    client = _main_client
    # encoded_message = message.encode('utf-8')
    encoded_message = _encrypt_public(message)
    # print('CLIENT: ENCRYPTED:', encoded_message)
    client.sendall(encoded_message)
    pprint(f"Sent:     {message}")
    # incoming_data = full_recieve(client)
    if recieve:
        # incoming_data = client.recv(1024).decode('utf-8')
        incoming_data = client.recv(1024)
        incoming_data = _decrypt_private(incoming_data)
        pprint(f"Received: {incoming_data}")
        return incoming_data

# def send_file(file, recieve: bool = False) -> any:
#     '''
#     A general tool function for sending files to the recipient (server, other client, etc)
#     '''
#     client = _main_client
#     encoded_file =

# def target_client(client_ip: str, client_port: int, mode: str) -> bool:
#     '''
#     Takes in the target clients ip and port, and will attempt to connect to them. If this fails, 
#     then it is possible the other client is not available.
#     This function returns a boolean, telling you whether it worked or not.
#     '''
#     global _HOST, _PORT, _valid_ports
#     global _main_client
#     global _do_print
#     # reset main_client
#     _main_client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
#     # overwrite host
#     _HOST = client_ip
#     # overwrite valid ports list
#     override_ports([client_port])
#     # overwrite port
#     _PORT = _valid_ports[0]
#     # setup other vars
#     save = _do_print

#     for i in range(30):
#         print(f'attempt {i}')
#         disable_print()
#         _main_client, _PORT = _cycle_port(_main_client)
#         _do_print = save
#         if _connected == True:
#             return True
#         time.sleep(1)
#     return False



# function for shutting down the client
def shutdown_client() -> bool:
    '''
    A function to shut down the client: returns a bool telling you whether it worked or not.
    '''
    global _main_client
    try:
        send_msg('end_session')
    except:
        pass
    try:
        _main_client.close()
        pprint('[CLIENT SHUTDOWN] Shutting down client...')
        return True
    except:
        return False

def start_client(connection_ip: str) -> bool:
    '''
    Starts the connection to the server, taking in an IP. 
    This function returns a bool, telling you whether it worked or not.
    '''
    global _main_client, _valid_ports, _PORT, _HOST
    _HOST = connection_ip

    # reset _main_client
    _main_client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    # overrides
    if len(_ov_ports) > 0:
        _valid_ports = _ov_ports
        _PORT = _valid_ports[0]
        pprint(f'[OVERRIDE] Overrided ports to: {_valid_ports}')
    
    # establish the connection to a port that the server is on
    _main_client, _PORT, state = _cycle_port(_main_client)
    return state


def start_client_listener() -> None:
    '''
    A function that acts on the recieving end to recieve information from the server. It adds to a message queue that can be accessed to retrieve data. 
    The latest entry in the data queue is the newest information. The queue gets cleared after 25 entries for efficiency and optimal storage.
    '''
    send_msg('listener')
    threading.Thread(target=lambda:_internal_client_listener()).start() # actual listener
    threading.Thread(target=lambda:_inbound_data_parser()).start() # parser to get individual commands
    threading.Thread(target=lambda:_parsed_data_decrypter()).start() # decrypts each command, goes into message queue
    # threading.Thread(target=lambda:_message_queue_cleaner()).start() # cleaner for list, saves latest val

def _message_queue_cleaner():
    global message_queue
    while True:
        if len(message_queue) > 25: # currently hard set value
            try:
                latest = message_queue[-1]
                # latest = message_queue[-5:]
                # print('saving latest:', latest)
                message_queue = [latest]
                # message_queue = latest
                # print('set new message queue to:', message_queue)
                # message_queue = []
                # print('cleaning data queue')
            except Exception as e:
                # print('clean error:', e)
                pass

#######################################################################################

_inbound_data = []
_parsed_data = [] 
# this function is responsible for decrypting and finally outputting data to the message queue
def _parsed_data_decrypter():
    while True:
        for parsed in _parsed_data:
            # print('[DECRYPTER] Decrypting:', parsed)
            try:
                decoded = _decrypt_private(parsed)
                # print('NEW DECRYPTED:', decoded)
                message_queue.append(decoded)
            except Exception as e:
                print('NEW DECRYPTION LOST:', e)
                # packet loss
                pass
            _parsed_data.remove(parsed)
            # exit()

# this function creates a parser which isolates each message
def _inbound_data_parser():
    while True:
        for data in _inbound_data:
            decoded_data = data.decode()
            # print('[PARSER] Parsing:', decoded_data)
            split_entry = [str(i) for i in decoded_data]
            new_segment = ''
            for letter in split_entry:
                if letter == '~':
                    location = split_entry.index(letter)
                    try:
                        next = split_entry[location + 1]
                        if next == 'E':
                            # new_segment = new_segment.encode()
                            new_segment = eval(new_segment)
                            # print('NEWS SEGMENT:', new_segment)
                            _parsed_data.append(new_segment)
                            new_segment = ''
                            break
                    except Exception as e:
                        print('NEW SEGMENT LOST:', e)
                        # cant go farther, lost packet
                        pass
                else:
                    new_segment += letter
            _inbound_data.remove(data)
            # exit()

# this function is responsible for taking in messages
def _internal_client_listener():
    global message_queue
    while True:
        ## OLD CODE
        # data = _main_client.recv(1024)
        # try:
        #     decoded = _decrypt_private(data)
        # except Exception as e:
        #     # print('packet loss error:', e)
        #     continue # packet loss
        # if decoded == '000':
        #     continue
        # # print('adding message to queue:', len(message_queue) + 1)
        # # print('recieved packet:', decoded)
        # message_queue.append(decoded)
        # # print('message queue:', message_queue)

        ## NEW CODE
        data = _main_client.recv(1024)
        _inbound_data.append(data)
        # print('[LISTENER] Recieving:', data)
        # exit()


'''
This is the comm file, which manages the sockets and communication.
The encryption tools is in the sec.py file, which can be accessed with the "sec" class object. 
The parser tools is in the parser.py file, which can be accessed with the "parser" class object.
'''

# included modules
import socket
import random
import pickle
import threading
import time
import platform

# files
from .sec import Sec
from .parser import Parser
from .exceptions import Exceptions
from .constants import Constants
from .pool import Pool

####################################################

# TODO ADD LOGGER

# this is the main class for the connection
class Comm:
    def __init__(self, host: tuple[str, int] = ['', 56774], UDP_bind: bool = False):
        # this is an instance of the security manager
        self.sec = Sec()
        # this is an instance of the parser
        self.parser = Parser()
        # this is the internal client used for sending and recieving
        if host[0]: self.host = host[0]
        else: self.host = socket.gethostbyname(socket.gethostname())
        self.port = host[1]
        # this is the target info
        self.target = (None, None)
        # this is the actual connected target info (FOR TCP)
        self.actual_target = (None, None)
        # this is the client for UDP for finding out who goes first
        self.UDP_client = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) # UDP
        # this is the main client for communication
        self.TCP_client = socket.socket(socket.AF_INET, socket.SOCK_STREAM) # TCP
        # this represents a dictionary of event queues
        self.eventRegistry = {}
        # this represents the connection type for when errors occur
        self.connectionType = Constants.ConnectionType.EVENT
        # this is the thread for the recieving function
        self.recvThread = threading.Thread(target=lambda:self.recv(), daemon=True)
        # this represents the system type
        self.systemType = platform.system()
        # this represents if the UDP client is binded or not
        self.UDP_binded = False
        # these are booleans for stopping threads
        self.stopRecv = False
        # this is a lock that, while something is sending, other things can not send
        self.sendLock = False
        # this represents if we have an active connected
        self.connected = False
        ###########################################################
        # if UDP_bind, immediately bind to host and port
        if UDP_bind: 
            self._bind_UDP()
            self.UDP_binded = True


    # this regenerates the UDP client, making a new object
    def _regen_UDP(self) -> None: self.UDP_client = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)


    # this regenerates the TCP client, making a new object
    def _regen_TCP(self) -> None: self.TCP_client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)


    # this binds the UDP client to the host machines ip and port
    def _bind_UDP(self) -> None: self.UDP_client.bind((self.host, self.port))


    # this binds the TCP client to the host machines ip and port
    def _bind_TCP(self) -> None: self.TCP_client.bind((self.host, self.port))


    # this returns the host IP and port in a tuple
    def get_host(self) -> tuple[str, int]: return self.host, self.port


    # this returns the actual target
    # that target being the active TCP connection, not the initial IP and port
    # before connectin
    def get_actual_target(self) -> tuple[str, int]: return self.actual_target
    

    # this starts the recv thread
    # for recieving messages and triggering events
    def start_recv(self) -> None: self.recvThread.start() if not self.recvThread.is_alive() else None


    # this function configures heartbeat things for the client
    # such as when to send them, how long to wait between each one, and how many to send
    def config_heartbeat(self, inactive_delay: int = 60, probe_interval: int = 10, probe_count: int = 5) -> None:
        self.TCP_client.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        if self.systemType == 'Windows': # Windows-specific options
            keepalive = (1, inactive_delay * 1000, probe_interval * 1000) # On, idle time (ms), interval (ms)
            self.TCP_client.ioctl(socket.SIO_KEEPALIVE_VALS, keepalive)
        elif self.systemType in ('Linux', 'Darwin'): # Linux/macOS-specific options
            self.TCP_client.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, inactive_delay) # Idle time before sending probes (in seconds)
            self.TCP_client.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, probe_interval) # Interval between probes (in seconds)
            self.TCP_client.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, probe_count) # Number of failed probes before closing
    

    # this function manages what happens when connection goes wrong,
    # and a connection is closing - typically with an error
    def _handle_close(self, error: Exception | None = None) -> None:
        # since we know an error happened and the connection likely is 
        # closed, we can force a close 
        if self.connected: self.close_connection(force=True)
        # handle the error according to how client is configured
        if self.connectionType == Constants.ConnectionType.EVENT: self._trigger(Constants.Event.ON_CLOSE, error)
        elif self.connectionType == Constants.ConnectionType.ERROR: raise error


    # this function runs the given events when requested
    # events are created using decorators
    def _trigger(self, eventType: Constants.Event, data) -> None:
        # run every function set up under the event
        try:
            for func in self.eventRegistry[eventType]: 
                func(data)
        # if no functions then there will be a key error, this is fine
        # so we can ignore
        except KeyError: return


    # this function handles the UDP connection that helps make the TCP connection
    # as well as the handshake, and the overall connection setup
    def TCP_connect(self, 
                     target_ip: str, 
                     target_port: int, 
                     timeout: int = 10,
                     attempts: int = 6,
                     connection_bias: Constants.ConnectionBias = Constants.ConnectionBias.NONE
                     ) -> None:
        # set target machine data
        self.target = (target_ip, target_port)
        # determine whether or not to use UDP
        if connection_bias != Constants.ConnectionBias.NONE: 
            # UDP is only used to determine who goes first / second
            # so if we can determine if we are not using it by the connection bias
            first = connection_bias
        else:
            # we use UDP to get the random number
            ourRandom, targetRandom = self._UDP_connect(timeout, attempts)
            # if True meaning we connect, they recv
            # if False, we recv and they connect
            first = ourRandom > targetRandom
        # we then find out whether to bind our TCP
        # or try to connect to the other end
        self._regen_TCP()
        if first:
            self.TCP_client.connect(self.target)
            self.actual_target = self.target
        else:
            # we try (attempts) times to connect
            # an invalid connection is if the client that connects
            # is not the one we wanted to connect to
            connectionSuccess = False
            for _ in range(attempts):
                self._regen_TCP()
                self._bind_TCP()
                self.TCP_client.listen(1) # only listen for 1 connection
                self.TCP_client, connectedTarget = self.TCP_client.accept()
                if connectedTarget[0] == self.target[0]: # verify IP, not port
                    self.actual_target = connectedTarget
                    connectionSuccess = True
                    break
            # raise error if connection failed
            if not connectionSuccess: raise Exceptions.ConnectionFailedError(f'Failed to connect to target machine (TCP) (attempts:{attempts})') 
        # set up the settings for heartbeat pings
        self.config_heartbeat()
        # do the handshake to exchange RSA keys
        self._handshake(first)
        self.connected = True
        # trigger connect event
        self._trigger(Constants.Event.ON_CONNECT, True)


    # this function manages finding out who goes first with making a TCP connection
    # and also who is first with exchanging RSA keys
    def _UDP_connect(self, timeout, attempts) -> tuple[int, int]:
        # first, we bind to our port / ip if not already
        if not self.UDP_binded: 
            self._regen_UDP()
            self._bind_UDP()
            self.UDP_binded = True
        # now, we generate and send a random number
        randNum = random.randint(0, 1000) + random.randint(0, 1000)
        # we try "attempts" times to connect and wait "timeout" seconds for a response
        connectionSuccess = False
        self.UDP_client.settimeout(timeout)
        for _ in range(attempts):
            try:
                # if we send the data and get data back,
                # then it succeeded
                self.UDP_client.sendto(str(randNum).encode(), self.target)
                data, self.target = self.UDP_client.recvfrom(1024)
                self.UDP_client.sendto(str(randNum).encode(), self.target) # make sure data got through
                # we decode the incoming value to make sure the two values aren't equal
                # if they are, we raise error (the chances are very low for this to happen)
                incomingNum = int(data.decode())
                if incomingNum == randNum: raise Exceptions.ConnectionFailedError('Role number generations were equal.')
                # otherwise connection was a success, break
                connectionSuccess = True
                break
            except TimeoutError: continue
        # if no success, raise error
        if not connectionSuccess: 
            raise Exceptions.ConnectionFailedError(f'Failed to connect to target machine (UDP) (attempts:{attempts})') 
        # we close our UDP and return the two numbers
        self.UDP_client.close()
        return (randNum, incomingNum)
    

    # this function manages handshakes for exchanging RSA keys
    # which are exchanged just to exchange symmetrical Fernet keys
    def _handshake(self, is_first: bool) -> None:
        if is_first:
            # we send our public RSA key
            self.TCP_client.sendall(pickle.dumps(self.sec.int_pub_key))
            # then recieve their public RSA key
            recievedPubKey = self.TCP_client.recv(1024)
            self.sec.load_RSA(pickle.loads(recievedPubKey))
            # now we send our Fernet key for actual encryption
            # since we are first, we don't need to recieve 
            # since the keys are the same
            encryptedFernet = self.sec.RSA_encrypt(self.sec.fernet_key)
            self.TCP_client.sendall(encryptedFernet)
        else:
            # we recieve their public key
            recievedPubKey = self.TCP_client.recv(1024)
            self.sec.load_RSA(pickle.loads(recievedPubKey))
            # then send our public key
            self.TCP_client.sendall(pickle.dumps(self.sec.int_pub_key))
            # now we recieve the other ends symmetrical token for actual encryption
            # since we are second
            encryptedFernet = self.TCP_client.recv(1024)
            self.sec.load_Fernet(self.sec.RSA_decrypt(encryptedFernet))
    

    # this function closes the connection between the two machines
    # gracefully :3
    def close_connection(self, force: bool = False) -> None: 
        if not self.connected: return
        self.stopRecv = True
        # this shuts down the read and write pipes gracefully
        # making sure that all data is recieved and sent properly
        # before closing
        if not force: self.TCP_client.shutdown(socket.SHUT_RDWR)
        self.TCP_client.close()
        self._regen_UDP()
        self._regen_TCP()
        self.UDP_binded = False
        self.connected = False
    

    # this is a function to send data to the other machine
    def send(self, data: any, ignore_errors: bool = False, lock_timeout: float = 10.0) -> None:
        # raise error message if data is empty and ignore is disabled,
        # otherwise return
        if len(data) == 0 or data is None:
            if ignore_errors: return
            raise Exceptions.EmptyDataError()
        # raise error if not connected
        if not self.connected: 
            if ignore_errors: return
            raise Exceptions.ClientNotConnectedError()
        # wait until no longer a lock on sending or if we reach the lock timeout
        timeout = time.time() + lock_timeout
        while self.sendLock:
            if time.time() >= timeout:
                if ignore_errors: return
                raise Exceptions.SendingTimeoutError()
        # set the send lock to True so that any other calls to send
        # data have to wait to send their data
        self.sendLock = True
        messageObject = Pool.Message(data) # create message object
        encryptedMessage = self.sec.Fernet_encrypt(messageObject) # encrypt data
        paddedMessage = self.parser.addPadding(encryptedMessage) # pad data
        try: 
            self.TCP_client.sendall(paddedMessage) # send actual data
            self.sendLock = False
        except ConnectionResetError as e: # other machine quit
            self.sendLock = False
            self._handle_close(e)
        except BrokenPipeError as e: # connection was properly closed and sending data was attempted
            self.sendLock = False
            self._handle_close(e)


    # this is a recieving function for recieving data
    def recv(self) -> None:
        while True:
            try: recieved = self.TCP_client.recv(1024)
            except ConnectionResetError as e: # other machine quit
                self._handle_close(e)
                return
            except ConnectionAbortedError as e: # host client closed
                self._handle_close(e)
                return
            # catch any other error that happens
            # if we are meant to exit thread, just ignore error and exit
            # otherwise raise the same exception again
            except Exception as e:
                if self.stopRecv: return
                raise e
            # if recieved is empty, then we got an EOF meaning the other socket
            # shutdown
            if not recieved: 
                self._handle_close(EOFError('EOF detected: Remote socket shutdown.'))
                return
            # remove padding from the recieved data
            unpadded = self.parser.removePadding(recieved)
            for indiv in unpadded:
                decrypted: Pool.Message = self.sec.Fernet_decrypt(indiv)
                decrypted.recieved_at = Pool._Tools._format_time()
                self._trigger(Constants.Event.ON_MESSAGE, decrypted)